"""
:class:`~pyobs.object.Object` is the base for almost all classes in *pyobs*. It adds some convenience methods
and helper methods for creating other Objects.

There are a few convenience functions:

    - :func:`~pyobs.object.create_object` creates objects from dictionaries.
    - :func:`~pyobs.object.get_object` is a wrapper around :func:`pyobs.object.create_object` that can do further checks.
    - :func:`~pyobs.object.get_safe_object` is a wrapper around :func:`~pyobs.object.get_object` that never raises
      exceptions.
"""

from __future__ import annotations

import asyncio
import copy
import datetime
import inspect
from collections import Coroutine
from typing import Union, Callable, TypeVar, Optional, Type, List, Tuple, Dict, Any, overload, TYPE_CHECKING, Awaitable
import logging
import pytz
from astroplan import Observer
from astropy.coordinates import EarthLocation

from pyobs.comm import Comm
from pyobs.comm.dummy import DummyComm

if TYPE_CHECKING:
    from pyobs.vfs import VirtualFileSystem

log = logging.getLogger(__name__)


"""Class of an Object."""
ObjectClass = TypeVar("ObjectClass")


"""Class of a proxy."""
ProxyType = TypeVar("ProxyType")


@overload
def get_object(
    config_or_object: Union[Dict[str, Any], ObjectClass, Type[ObjectClass], Any],
    object_class: Type[ObjectClass],
    **kwargs: Any,
) -> ObjectClass:
    ...


@overload
def get_object(
    config_or_object: Union[Dict[str, Any], ObjectClass, Type[ObjectClass], Any], object_class: None, **kwargs: Any
) -> Any:
    ...


def get_object(
    config_or_object: Union[Dict[str, Any], ObjectClass, Type[ObjectClass], Any],
    object_class: Optional[Type[ObjectClass]] = None,
    **kwargs: Any,
) -> Union[ObjectClass, Any]:
    """Creates object from config or returns object directly, both optionally after check of type.

    Args:
        config_or_object: A configuration dict or an object itself to create/check. If a dict with a class key
            is given, a new object is created.
        object_class: Class to check object against.

    Returns:
        (New) object (created from config) that optionally passed class check.

    Raises:
        TypeError: If the object does not match the given class.
    """

    if config_or_object is None:
        raise TypeError("No config or object given.")

    elif isinstance(config_or_object, dict):
        # copy kwargs to config_or_object, so that we don't have any duplicates
        for k, v in kwargs.items():
            config_or_object[k] = v

        # a dict is given, so create object
        obj = create_object(config_or_object)

    elif inspect.isclass(config_or_object):
        # config_or_object is a type, so create it using its constructor
        obj = config_or_object(**kwargs)

    else:
        # just use given object
        obj = config_or_object

    # do we need a type check and does the given object pass?
    if object_class is not None and not isinstance(obj, object_class):
        raise TypeError("Provided object is not of requested type %s." % object_class.__name__)
    return obj


@overload
def get_safe_object(
    config_or_object: Union[ObjectClass, Dict[str, Any]], object_class: Type[ObjectClass], **kwargs: Any
) -> Optional[ObjectClass]:
    ...


@overload
def get_safe_object(config_or_object: Union[ObjectClass, Any], object_class: None, **kwargs: Any) -> Optional[Any]:
    ...


def get_safe_object(
    config_or_object: Union[Dict[str, Any], Any], object_class: Optional[Type[ObjectClass]] = None, **kwargs: Any
) -> Optional[Union[ObjectClass, Any]]:
    """Calls get_object in a safe way and returns None, if an exceptions thrown.

    Args:
        config_or_object: A configuration dict or an object itself to create/check. If a dict with a class key
            is given, a new object is created.
        object_class: Class to check object against.

    Returns:
        (New) object (created from config) that optionally passed class check or None.
    """
    try:
        return get_object(config_or_object, object_class, **kwargs)
    except Exception:
        return None


def get_class_from_string(class_name: str) -> Type[Any]:
    """Get class from a given string.

    Args:
        class_name: Name of class as string.

    Returns:
        Actual class.
    """

    parts = class_name.split(".")
    module_name = ".".join(parts[:-1])
    cls: Type[Any] = __import__(module_name)
    for comp in parts[1:]:
        cls = getattr(cls, comp)
    return cls


def create_object(config: Dict[str, Any], *args: Any, **kwargs: Any) -> Any:
    """Create object from dict config.

    Args:
        config: Config to create object from
        *args: Parameters to be passed to object.
        **kwargs: Parameters to be passed to object.

    Returns:
        Created object.
    """

    # get class name
    class_name = config["class"]

    # create class
    klass = get_class_from_string(class_name)

    # remove class from kwargs
    cfg = copy.copy(config)
    del cfg["class"]

    # create object
    return klass(*args, **cfg, **kwargs)


class Object:
    """Base class for all objects in *pyobs*."""

    def __init__(
        self,
        vfs: Optional[Union["VirtualFileSystem", Dict[str, Any]]] = None,
        comm: Optional[Union[Comm, Dict[str, Any]]] = None,
        timezone: Union[str, datetime.tzinfo] = "utc",
        location: Optional[Union[str, Dict[str, Any], EarthLocation]] = None,
        observer: Optional[Observer] = None,
        **kwargs: Any,
    ):
        """
        .. note::

            Objects must always be opened and closed using :meth:`~pyobs.object.Object.open` and
            :meth:`~pyobs.object.Object.close`, respectively.

        This class provides a :class:`~pyobs.vfs.VirtualFileSystem`, a timezone and a location. From the latter two, an
        observer object is automatically created.

        Object also adds support for easily adding threads using the :meth:`~pyobs.object.Object.add_background_task`
        method as well as a watchdog thread that automatically restarts threads, if requested.

        Using :meth:`~pyobs.object.Object.add_child_object`, other objects can be (created an) attached to this object,
        which then automatically handles calls to :meth:`~pyobs.object.Object.open` and :meth:`~pyobs.object.Object.close`
        on those objects.

        Args:
            vfs: VFS to use (either object or config)
            comm: Comm object to use
            timezone: Timezone at observatory.
            location: Location of observatory, either a name or a dict containing latitude, longitude, and elevation.

        """
        from pyobs.vfs import VirtualFileSystem

        # child objects
        self._child_objects: List[Any] = []

        # create vfs
        if vfs:
            self.vfs = get_object(vfs, VirtualFileSystem)
        else:
            self.vfs = VirtualFileSystem()

        # timezone
        if isinstance(timezone, datetime.tzinfo):
            self.timezone = timezone
        elif isinstance(timezone, str):
            self.timezone = pytz.timezone(timezone)
        else:
            raise ValueError("Unknown format for timezone.")

        # location
        if location is None:
            self.location = None
        elif isinstance(location, EarthLocation):
            self.location = location
        elif isinstance(location, str):
            self.location = EarthLocation.of_site(location)
        elif isinstance(location, dict):
            self.location = EarthLocation.from_geodetic(
                location["longitude"], location["latitude"], location["elevation"]
            )
        else:
            raise ValueError("Unknown format for location.")

        # create observer
        self.observer = observer
        if self.observer is None and self.location is not None and self.timezone is not None:
            log.info(
                "Setting location to longitude=%s, latitude=%s, and elevation=%s.",
                self.location.lon,
                self.location.lat,
                self.location.height,
            )
            self.observer = Observer(location=self.location, timezone=timezone)

        # comm object
        self.comm: Comm
        if comm is None:
            self.comm = DummyComm()
        elif isinstance(comm, Comm):
            self.comm = comm
        elif isinstance(comm, dict):
            log.info("Creating comm object...")
            self.comm = get_object(comm, Comm)
        else:
            raise ValueError("Invalid Comm object")

        # opened?
        self._opened = False

        # background tasks
        self._background_tasks: Dict[
            Callable[..., Coroutine[Any, Any, None]], Tuple[Optional[asyncio.Task[bool]], bool]
        ] = {}
        self._watchdog_task: Optional[asyncio.Task[None]] = None

    def add_background_task(self, func: Callable[..., Coroutine[Any, Any, None]], restart: bool = True) -> None:
        """Add a new function that should be run in the background.

        MUST be called in constructor of derived class or at least before calling open() on the object.

        Args:
            func: Func to add.
            restart: Whether to restart this function.
        """

        # create thread
        self._background_tasks[func] = (None, restart)

    async def open(self) -> None:
        """Open module."""

        # start background tasks
        for func, (task, restart) in self._background_tasks.items():
            log.info("Starting background task for %s...", func.__name__)
            self._background_tasks[func] = (asyncio.create_task(self._background_func(func)), restart)
        if len(self._background_tasks) > 0:
            self._watchdog_task = asyncio.create_task(self._watchdog())

        # open child objects
        for obj in self._child_objects:
            if hasattr(obj, "open"):
                if asyncio.iscoroutinefunction(obj.open):
                    await obj.open()
                else:
                    obj.open()

        # success
        self._opened = True

    @property
    def opened(self) -> bool:
        """Whether object has been opened."""
        return self._opened

    async def close(self) -> None:
        """Close module."""

        # close child objects
        for obj in self._child_objects:
            if hasattr(obj, "close"):
                await obj.close()

        # join watchdog and then all threads
        if self._watchdog_task and not self._watchdog_task.done():
            self._watchdog_task.cancel()
        for func, (task, restart) in self._background_tasks.items():
            if task and not task.done():
                task.cancel()

    @staticmethod
    async def _background_func(target: Callable[..., Coroutine[Any, Any, None]]) -> None:
        """Run given function.

        Args:
            target: Function to run.
        """
        try:
            await target()

        except asyncio.CancelledError:
            # task was canceled
            return

        except:
            log.exception("Exception in thread method %s." % target.__name__)

    def quit(self) -> None:
        """Can be overloaded to quit program."""
        pass

    async def _watchdog(self) -> None:
        """Watchdog thread that tries to restart threads if they quit."""

        while True:
            # get dead taks that need to be restarted
            dead = {}
            for func, (task, restart) in self._background_tasks.items():
                if task.done():
                    dead[func] = restart

            # restart dead tasks or quit
            for func, restart in dead.items():
                if restart:
                    log.error("Background task for %s has died, restarting...", func.__name__)
                    del self._background_tasks[func]
                    self._background_tasks[func] = (asyncio.create_task(self._background_func(func)), restart)
                else:
                    log.error("Background task for %s has died, quitting...", func.__name__)
                    self.quit()
                    return

            # sleep a little
            await asyncio.sleep(1)

    @overload
    def get_object(
        self,
        config_or_object: Union[Dict[str, Any], ObjectClass, Type[ObjectClass], Any],
        object_class: Type[ObjectClass],
        copy_comm: bool = True,
        **kwargs: Any,
    ) -> ObjectClass:
        ...

    @overload
    def get_object(
        self, config_or_object: Type[ObjectClass], object_class: None = None, copy_comm: bool = True, **kwargs: Any
    ) -> ObjectClass:
        ...

    @overload
    def get_object(
        self,
        config_or_object: Union[Dict[str, Any], ObjectClass, Type[ObjectClass], Any],
        object_class: None,
        copy_comm: bool = True,
        **kwargs: Any,
    ) -> Any:
        ...

    def get_object(
        self,
        config_or_object: Union[Dict[str, Any], ObjectClass, Type[ObjectClass], Any],
        object_class: Optional[Type[ObjectClass]] = None,
        copy_comm: bool = True,
        **kwargs: Any,
    ) -> Union[ObjectClass, Any]:
        """Creates object from config or returns object directly, both optionally after check of type.

        Args:
            config_or_object: A configuration dict or an object itself to create/check. If a dict with a class key
                is given, a new object is created.
            object_class: Class to check object against.
            copy_comm: Copy comm from this object to the new one.

        Returns:
            (New) object (created from config) that optionally passed class check.

        Raises:
            TypeError: If the object does not match the given class.
        """

        # set parameters
        params = copy.copy(kwargs)
        params.update(
            {"timezone": self.timezone, "location": self.location, "vfs": self.vfs, "observer": self.observer}
        )
        if copy_comm:
            params["comm"] = self.comm

        # get it
        return get_object(config_or_object, object_class, **params)

    @overload
    def get_safe_object(
        self,
        config_or_object: Union[Dict[str, Any], ObjectClass, Type[ObjectClass], Any],
        object_class: Type[ObjectClass],
        copy_comm: bool = True,
        **kwargs: Any,
    ) -> Optional[ObjectClass]:
        ...

    @overload
    def get_safe_object(
        self,
        config_or_object: Union[Dict[str, Any], ObjectClass, Type[ObjectClass], Any],
        object_class: None,
        copy_comm: bool = True,
        **kwargs: Any,
    ) -> Optional[Any]:
        ...

    def get_safe_object(
        self,
        config_or_object: Union[Dict[str, Any], ObjectClass, Type[ObjectClass], Any],
        object_class: Optional[Type[ObjectClass]] = None,
        copy_comm: bool = True,
        **kwargs: Any,
    ) -> Optional[Union[ObjectClass, Any]]:
        """Calls get_object in a safe way and returns None, if an exceptions thrown."""
        try:
            return self.get_object(config_or_object, object_class=object_class, copy_comm=copy_comm, **kwargs)
        except Exception:
            log.exception("test")
            return None

    @overload
    def add_child_object(
        self,
        config_or_object: Union[Dict[str, Any], ObjectClass, Type[ObjectClass], Any],
        object_class: Type[ObjectClass],
        copy_comm: bool = True,
        **kwargs: Any,
    ) -> ObjectClass:
        ...

    @overload
    def add_child_object(
        self, config_or_object: Type[ObjectClass], object_class: None = None, copy_comm: bool = True, **kwargs: Any
    ) -> ObjectClass:
        ...

    @overload
    def add_child_object(
        self,
        config_or_object: Union[Dict[str, Any], ObjectClass, Type[ObjectClass], Any],
        object_class: None,
        copy_comm: bool = True,
        **kwargs: Any,
    ) -> Any:
        ...

    def add_child_object(
        self,
        config_or_object: Union[Dict[str, Any], ObjectClass, Type[ObjectClass], Any],
        object_class: Optional[Type[ObjectClass]] = None,
        copy_comm: bool = True,
        **kwargs: Any,
    ) -> Union[ObjectClass, Any]:
        """Create a new sub-module, which will automatically be opened and closed.

        Args:
            config_or_object: Module definition
            object_class: Class for new module
            copy_comm: Copy comm from this object to the new one.

        Returns:
            The created module.
        """

        # get object
        obj = self.get_object(config_or_object, object_class=object_class, copy_comm=copy_comm, **kwargs)

        # add to list
        self._child_objects.append(obj)

        # return it
        return obj

    @overload
    async def proxy(self, name_or_object: Union[str, object], obj_type: Type[ProxyType]) -> ProxyType:
        ...

    @overload
    async def proxy(self, name_or_object: Union[str, object], obj_type: Optional[Type[ProxyType]] = None) -> Any:
        ...

    async def proxy(
        self, name_or_object: Union[str, object], obj_type: Optional[Type[ProxyType]] = None
    ) -> Union[Any, ProxyType]:
        """Returns object directly if it is of given type. Otherwise get proxy of client with given name and check type.

        If name_or_object is an object:
            - If it is of type (or derived), return object.
            - Otherwise raise exception.
        If name_name_or_object is string:
            - Create proxy from name and raise exception, if it doesn't exist.
            - Check type and raise exception if wrong.
            - Return object.

        Args:
            name_or_object: Name of object or object itself.
            obj_type: Expected class of object.

        Returns:
            Object or proxy to object.

        Raises:
            ValueError: If proxy does not exist or wrong type.
        """
        return await self.comm.proxy(name_or_object, obj_type)


__all__ = ["get_object", "get_class_from_string", "create_object", "Object"]



from typing import Tuple, Dict, List, Optional, Any

import numpy as np
import logging
from scipy import ndimage

from pyobs.utils.focusseries.base import FocusSeries
from pyobs.utils.curvefit import fit_hyperbola
from pyobs.images import Image


log = logging.getLogger(__name__)


class ProjectionFocusSeries(FocusSeries):
    """Focus series from projected x/y."""

    __module__ = "pyobs.utils.focusseries"

    def __init__(self, backsub: bool = True, xbad: Optional[List[int]] = None, ybad: Optional[List[int]] = None):
        """Initialize a new projection focus series.

        Args:
            backsub: Do background subtraction?
            xbad: Bad rows
            ybad: Bad columns
        """

        # test imports

        # stuff
        self._backsub = backsub
        self._xbad = xbad
        self._ybad = ybad
        self._data: List[Dict[str, float]] = []

    def reset(self) -> None:
        """Reset focus series."""
        self._data = []

    def analyse_image(self, image: Image, focus_value: float) -> None:
        """Analyse given image.

        Args:
            image: Image to analyse
            focus_value: Value to fit along, e.g. focus value or its offset
        """

        # clean data
        if image.data is None:
            return
        data = self._clean(image.data)

        # get projections
        xproj = np.mean(data, axis=0)
        yproj = np.mean(data, axis=1)
        nx = len(xproj)
        ny = len(yproj)

        # remove background gradient
        xclean = xproj - ndimage.uniform_filter1d(xproj, nx // 10)
        yclean = yproj - ndimage.uniform_filter1d(yproj, ny // 10)

        # get window functions
        xwind = self._window_function(xclean, border=3)
        ywind = self._window_function(yclean, border=3)

        # calculate correlation functions
        xavg = np.average(xclean)
        yavg = np.average(yclean)
        x = xwind * (xclean - xavg) / xavg
        y = ywind * (yclean - yavg) / yavg
        xcorr = np.correlate(x, x, mode="same")
        ycorr = np.correlate(y, y, mode="same")

        # filter out the peak (e.g. cosmics, ...)
        # imx = np.argmax(xcorr)
        # xcorr[imx] = 0.5 * (xcorr[imx - 1] + xcorr[imx + 1])
        # imx = np.argmax(ycorr)
        # ycorr[imx] = 0.5 * (ycorr[imx - 1] + ycorr[imx + 1])

        # fit cc functions to get fwhm
        xfit = self._fit_correlation(xcorr)
        yfit = self._fit_correlation(ycorr)

        # log it
        log.info(
            "Found x=%.1f+-%.1f and y=%.1f+-%.1f.",
            xfit.params["fwhm"].value,
            xfit.params["fwhm"].stderr,
            yfit.params["fwhm"].value,
            yfit.params["fwhm"].stderr,
        )

        # add to list
        self._data.append(
            {
                "focus": focus_value,
                "x": float(xfit.params["fwhm"].value),
                "xerr": float(xfit.params["fwhm"].stderr),
                "y": float(yfit.params["fwhm"].value),
                "yerr": float(yfit.params["fwhm"].stderr),
            }
        )

    def fit_focus(self) -> Tuple[float, float]:
        """Fit focus from analysed images

        Returns:
            Tuple of new focus and its error
        """

        # get data
        focus = [d["focus"] for d in self._data]
        xfwhm = [d["x"] for d in self._data]
        xsig = [d["xerr"] for d in self._data]
        yfwhm = [d["y"] for d in self._data]
        ysig = [d["yerr"] for d in self._data]

        # fit focus
        try:
            xfoc, xerr = fit_hyperbola(focus, xfwhm, xsig)
            yfoc, yerr = fit_hyperbola(focus, yfwhm, ysig)

            # weighted mean
            xerr = np.sqrt(xerr)
            yerr = np.sqrt(yerr)
            foc = (xfoc / xerr + yfoc / yerr) / (1.0 / xerr + 1.0 / yerr)
            err = 2.0 / (1.0 / xerr + 1.0 / yerr)
        except (RuntimeError, RuntimeWarning):
            raise ValueError("Could not find best focus.")

        # get min and max foci
        min_focus = np.min(focus)
        max_focus = np.max(focus)
        if foc < min_focus or foc > max_focus:
            raise ValueError("New focus out of bounds: {0:.3f}+-{1:.3f}mm.".format(foc, err))

        # return it
        return float(foc), float(err)

    @staticmethod
    def _window_function(arr: np.ndarray, border: int = 0) -> np.ndarray:
        """
        Creates a sine window function of the same size as some 1-D array "arr".
        Optionally, a zero border at the edges is added by "scrunching" the window.
        """
        ndata = len(arr)
        nwind = ndata - 2 * border
        w = np.zeros(ndata)
        for i in range(nwind):
            w[i + border] = np.sin(np.pi * (i + 1.0) / (nwind + 1.0))
        return w

    @staticmethod
    def _clean(
        data: np.ndarray, backsub: bool = True, xbad: Optional[np.ndarray] = None, ybad: Optional[np.ndarray] = None
    ) -> np.ndarray:
        """
        Removes global slopes and fills up bad rows (ybad) or columns (xbad).
        """
        (ny, nx) = data.shape

        # REMOVE BAD COLUMNS AND ROWS
        if xbad is not None:
            x1 = xbad - 1
            if x1 < 0:
                x1 = 1
            x2 = x1 + 2
            if x2 >= nx:
                x2 = nx - 1
                x1 = x2 - 2
            for j in range(ny):
                data[j][xbad] = 0.5 * (data[j][x1] + data[j][x2])
        if ybad is not None:
            y1 = ybad - 1
            if y1 < 0:
                y1 = 1
            y2 = y1 + 2
            if y2 >= ny:
                y2 = ny - 1
                y1 = y2 - 2
            for i in range(nx):
                data[ybad][i] = 0.5 * (data[y1][i] + data[y2][i])

        # REMOVE GLOBAL SLOPES
        if backsub:
            xsl = np.median(data, axis=0)
            ysl = np.median(data, axis=1).reshape((ny, 1))
            xsl -= np.mean(xsl)
            ysl -= np.mean(ysl)
            xslope = np.tile(xsl, (ny, 1))
            yslope = np.tile(ysl, (1, nx))
            return data - xslope - yslope
        else:
            return data

    @staticmethod
    def _fit_correlation(correl: np.ndarray) -> Any:
        from lmfit.models import GaussianModel

        # create Gaussian model
        model = GaussianModel()

        # initial guess
        x = np.arange(len(correl))
        pars = model.guess(correl, x=x)
        pars["sigma"].value = 20.0

        # fit
        return model.fit(correl, pars, x=x)


__all__ = ["ProjectionFocusSeries"]



from __future__ import annotations
import logging
from typing import Union, cast, Optional, Any

from pyobs.interfaces import IBinning, IWindow, IFilters, IImageGrabber
from pyobs.modules import Module

log = logging.getLogger(__name__)


class CameraSettingsMixin:
    """Mixin for a device that should be able to set camera settings."""

    __module__ = "pyobs.mixins"

    def __init__(
        self,
        filters: Optional[Union[str, IFilters]] = None,
        filter_name: Optional[str] = None,
        binning: Optional[int] = None,
        **kwargs: Any,
    ):
        """Initializes the mixin.

        Args:
            filters: Filter wheel module.
            filter: Filter to set.
            binning: Binning to set.
        """

        # store
        self.__camerasettings_filters = filters
        self.__camerasettings_filter = filter_name
        self.__camerasettings_binning = binning

    async def _do_camera_settings(self, camera: Union[Module, IImageGrabber, IFilters, IBinning, IWindow]) -> None:
        """Do camera settings for given camera."""

        # check type
        if not isinstance(self, Module) or not isinstance(self, CameraSettingsMixin):
            raise ValueError("This is not a module")

        # filter
        if self.__camerasettings_filters is not None and self.__camerasettings_filter is not None:
            # get proxy
            log.info("Getting proxy for filter wheel...")
            filters = await cast(Module, self).proxy(self.__camerasettings_filters, IFilters)

            # set it
            log.info("Setting filter to %s...", self.__camerasettings_filter)
            await filters.set_filter(self.__camerasettings_filter)

        # camera settings
        if self.__camerasettings_binning is not None and isinstance(camera, IBinning):
            log.info("Setting binning to %dx%d...", self.__camerasettings_binning, self.__camerasettings_binning)
            await camera.set_binning(self.__camerasettings_binning, self.__camerasettings_binning)
        if isinstance(camera, IWindow):
            log.info("Set window to full frame...")
            full_frame = await camera.get_full_frame()
            if full_frame is not None:
                await camera.set_window(*full_frame)
            else:
                raise ValueError("Could not get full frame size.")


__all__ = ["CameraSettingsMixin"]



import logging
from typing import List, Dict, Optional, Any, Tuple


log = logging.getLogger(__name__)


class FitsNamespaceMixin:
    """Mixin for IFitsHeaderProvider modules that filters FITS headers by namespace."""

    __module__ = "pyobs.mixins"

    def __init__(self, fits_namespaces: Optional[Dict[str, List[str]]] = None, **kwargs: Any):
        self.__namespaces = {} if fits_namespaces is None else fits_namespaces

    def _filter_fits_namespace(
        self, hdr: Dict[str, Tuple[Any, str]], sender: str, namespaces: Optional[List[str]] = None
    ) -> Dict[str, Tuple[Any, str]]:
        """Filter FITS header keywords by given namespaces. If no namespaces are given, let all through. Always
        let keywords with this module's name as namespace pass.

        Args:
            hdr: Input header to filter
            namespaces: Requested namespaces
            sender: Name of module that requested headers

        Returns:
            Filtered FITS header
        """

        # no namespaces?
        if not self.__namespaces:
            return hdr

        # get list of FITS headers that we let pass
        keywords: List[str] = []

        # is the sender name in my namespaces?
        if sender in self.__namespaces:
            # add namespace
            self.__add_namespace(sender, keywords, hdr)

        # loop all given namespaces
        if namespaces is not None:
            for name in namespaces:
                # does namespace exist in my config?
                if name in self.__namespaces:
                    # add namespace
                    self.__add_namespace(name, keywords, hdr)

        # make unique
        keywords = list(set(keywords))

        # return filtered header
        return {k: v for k, v in hdr.items() if k in keywords}

    def __add_namespace(self, name: str, keywords: List[str], hdr: Dict[str, Any]) -> None:
        """Add FITS header keywords from namespace to list of valid keywords

        Args:
            name: Name of namespace
            keywords: List of valid keywords, which will be added to
            hdr: Full unfiltered header
        """

        # what to add?
        if name not in self.__namespaces:
            # given namespace doesn't exist? then add none
            return
        elif self.__namespaces[name] is None:
            # take all keywords, if namespace is empty or none are given
            keywords.extend(hdr.keys())
        elif isinstance(self.__namespaces[name], list):
            # take only keywords from list
            keywords.extend(self.__namespaces[name])
        else:
            log.error("Unknown namespace format: %s", self.__namespaces[name])


__all__ = ["FitsNamespaceMixin"]



from abc import ABCMeta, abstractmethod
from typing import Any

from pyobs.images import Image
from pyobs.object import Object


class ImageProcessor(Object, metaclass=ABCMeta):
    def __init__(self, **kwargs: Any):
        """Init new image processor."""
        Object.__init__(self, **kwargs)

    @abstractmethod
    async def __call__(self, image: Image) -> Image:
        """Processes an image.

        Args:
            image: Image to process.

        Returns:
            Processed image.
        """
        ...

    async def reset(self) -> None:
        """Resets state of image processor"""
        pass


__all__ = ["ImageProcessor"]



from __future__ import annotations
from abc import ABCMeta, abstractmethod
from typing import Tuple, TYPE_CHECKING, Any, Optional, List, Dict

from pyobs.object import Object
from pyobs.robotic.scripts import Script
from pyobs.utils.time import Time

if TYPE_CHECKING:
    from pyobs.robotic.taskschedule import TaskSchedule
    from pyobs.robotic.taskrunner import TaskRunner
    from pyobs.robotic.taskarchive import TaskArchive


class Task(Object, metaclass=ABCMeta):
    @property
    @abstractmethod
    def id(self) -> Any:
        """ID of task."""
        ...

    @property
    @abstractmethod
    def name(self) -> str:
        """Returns name of task."""
        ...

    @property
    @abstractmethod
    def duration(self) -> float:
        """Returns estimated duration of task in seconds."""
        ...

    @property
    @abstractmethod
    def start(self) -> Time:
        """Start time for task"""
        ...

    @property
    @abstractmethod
    def end(self) -> Time:
        """End time for task"""
        ...

    @abstractmethod
    async def can_run(self, scripts: Optional[Dict[str, Script]] = None) -> bool:
        """Checks, whether this task could run now.

        Returns:
            True, if task can run now.
        """
        ...

    @property
    @abstractmethod
    def can_start_late(self) -> bool:
        """Whether this tasks is allowed to start later than the user-set time, e.g. for flatfields.

        Returns:
            True, if task can start late.
        """
        ...

    @abstractmethod
    async def run(
        self,
        task_runner: TaskRunner,
        task_schedule: Optional[TaskSchedule] = None,
        task_archive: Optional[TaskArchive] = None,
        scripts: Optional[Dict[str, Script]] = None,
    ) -> None:
        """Run a task"""
        ...

    @abstractmethod
    def is_finished(self) -> bool:
        """Whether task is finished."""
        ...

    def get_fits_headers(self, namespaces: Optional[List[str]] = None) -> Dict[str, Tuple[Any, str]]:
        """Returns FITS header for the current status of this module.

        Args:
            namespaces: If given, only return FITS headers for the given namespaces.

        Returns:
            Dictionary containing FITS headers.
        """
        return {}


__all__ = ["Task"]



from abc import ABCMeta, abstractmethod
from typing import Any

from .interface import Interface
from pyobs.utils.enums import ModuleState


class IModule(Interface, metaclass=ABCMeta):
    """The module is actually a module. Implemented by all modules."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def get_label(self, **kwargs: Any) -> str:
        """Returns label of module."""
        ...

    @abstractmethod
    async def get_version(self, **kwargs: Any) -> str:
        """Returns pyobs version of module."""
        ...

    @abstractmethod
    async def get_state(self, **kwargs: Any) -> ModuleState:
        """Returns current state of module."""
        ...

    @abstractmethod
    async def reset_error(self, **kwargs: Any) -> bool:
        """Reset error of module, if any."""
        ...

    @abstractmethod
    async def get_error_string(self, **kwargs: Any) -> str:
        """Returns description of error, if any."""
        ...


__all__ = ["IModule"]



from typing import List, Tuple
import matplotlib.pyplot as plt
import numpy as np


class SphericalGrid:
    """Methods for creating points on a spherical lon/lat grid (e.g. Az/Alt or RA/Dec."""

    @staticmethod
    def lonlat(n_lon: int, n_lat: int) -> List[Tuple[float, float]]:
        """Creates a grid with points at the intersections of longitudinal and latitudinal lines.

        Params:
            n_lon: Number of longitudinal points.
            n_lat: Number of latitudinal points.

        Returns:
            Lat/lon grid.
        """
        grid = []
        for lon in np.linspace(0, 360.0 - 360.0 / n_lon, n_lon):
            for lat in np.linspace(-90, 90, n_lat):
                grid.append((lon, lat))
        return grid

    @staticmethod
    def equidistributed(n: int) -> List[Tuple[float, float]]:
        """Creates equidistributed points on the surface of a sphere

        See https://www.cmu.edu/biolphys/deserno/pdf/sphere_equi.pdf

        Params:
            n:  Number of points

        Returns:
            Lat/lon grid.
        """

        # init
        a = 4 * np.pi / n
        d = np.sqrt(a)
        m_phi = round(np.pi / d)
        d_phi = np.pi / m_phi
        d_varphi = a / d_phi
        grid = []

        # conversion radians -> degrees
        r2d = 180.0 / np.pi

        # loop latitudinal
        for m in range(0, m_phi):
            lat = np.pi * (m + 0.5) / m_phi
            m_varphi = round(2 * np.pi * np.sin(lat) / d_varphi)

            # loop longitudinal
            for n in range(0, m_varphi):
                lon = 2 * np.pi * n / m_varphi

                # append to grid
                grid.append((lon * r2d, lat * r2d - 90.0))

        # finished
        return grid

    @staticmethod
    def convert_to_cartesian(grid: List[Tuple[float, float]], radius: float = 1.0) -> List[Tuple[float, float, float]]:
        """Convert a grid to cartesian coordinates.

        Params:
            grid: Grid to convert.
            radius: Radius of sphere.

        Returns:
            List of x/y/z points.
        """

        # conversion radians -> degrees
        r2d = 180.0 / np.pi

        # calculate x/y/z coordinates, assuming r=1
        return [
            (
                radius * np.cos(lat / r2d) * np.cos(lon / r2d),
                radius * np.cos(lat / r2d) * np.sin(lon / r2d),
                radius * np.sin(lat / r2d),
            )
            for lon, lat in grid
        ]

    @staticmethod
    def plot_cartesian(grid: List[Tuple[float, float, float]]) -> None:
        """Plot cartesian grid.

        Params:
            grid: Cartesian grid to plot.
        """
        fig = plt.figure()
        ax = fig.add_subplot(projection="3d")
        x, y, z = list(zip(*grid))
        ax.scatter(x, z, y)



import logging
import numpy as np
from typing import List, Union, Any, Optional, Dict
from astropy.wcs import WCS
from astropy.wcs.utils import proj_plane_pixel_scales

from pyobs.modules import Module
from pyobs.events import NewImageEvent, Event
from pyobs.utils.publisher import Publisher
from pyobs.utils.time import Time

log = logging.getLogger(__name__)


class Seeing(Module):
    """Measures seeing on reduced images with a catalog."""

    __module__ = "pyobs.modules.image"

    def __init__(
        self,
        sources: Optional[Union[str, List[str]]] = None,
        publisher: Optional[Union[Publisher, Dict[str, Any]]] = None,
        max_ellipticity: float = 0.2,
        correct_for_airmass: bool = True,
        **kwargs: Any,
    ):
        """Creates a new seeing estimator.

        Args:
            sources: List of sources (e.g. cameras) to process images from or None for all.
            publisher: Publisher to publish results to.
            max_ellipticity: Maximum ellipticity for sources to consider.
            correct_for_zenith: Whether to correct seeing for airmass.
        """
        Module.__init__(self, **kwargs)

        # stuff
        self._sources = [sources] if isinstance(sources, str) else sources
        self._publisher = self.add_child_object(publisher, Publisher)
        self._max_ellipticity = max_ellipticity
        self._correct_for_airmass = correct_for_airmass

    async def open(self) -> None:
        """Open module."""
        await Module.open(self)

        # subscribe to channel with new images
        log.info("Subscribing to new image events...")
        await self.comm.register_event(NewImageEvent, self.process_new_image_event)

    async def process_new_image_event(self, event: Event, sender: str) -> bool:
        """Puts a new images in the DB with the given ID.

        Args:
            event:  New image event
            sender: Who sent the event?

        Returns:
            Success
        """
        if not isinstance(event, NewImageEvent):
            return False

        # filter by source
        if self._sources is not None and sender not in self._sources:
            return False

        # put into queue
        log.info("Received new image event from %s.", sender)

        # download image
        try:
            log.info("Downloading file %s...", event.filename)
            image = await self.vfs.read_image(event.filename)

        except FileNotFoundError:
            log.error("Could not download image.")
            return False

        # get catalog
        cat = image.catalog
        if cat is None:
            # no catalog found in file
            return False

        # filter by ellipticity
        cat = cat[cat["ellipticity"] < self._max_ellipticity]

        # get WCS and pixel size
        wcs = WCS(image.header)
        pix_size = abs(proj_plane_pixel_scales(wcs)[0] * 3600.0)

        # calculate seeing
        seeing = np.mean(cat["fwhm"]) * pix_size

        # correct for airmass?
        if self._correct_for_airmass:
            # Seeing S as function of seeing S0 at zenith and airmass a:
            # S = S0 * a^0.6
            # see https://www.astro.auth.gr/~seeing-gr/seeing_gr_files/theory/node17.html
            # need airmass
            if "AIRMASS" in image.header:
                seeing /= image.header["AIRMASS"] ** 0.6
            else:
                # could not correct
                return False

        # log it
        if self._publisher is not None:
            await self._publisher(time=Time.now().isot, seeing=seeing)
        return True


__all__ = ["Seeing"]



import logging
from typing import Union, Any, Dict

from pyobs.object import get_object
from pyobs.images import Image
from pyobs.images.processors.exptime.exptime import ExpTimeEstimator
from pyobs.images.processors.detection import SourceDetection

log = logging.getLogger(__name__)


class StarExpTimeEstimator(ExpTimeEstimator):
    """Estimate exposure time from a star."""

    __module__ = "pyobs.images.processors.exptime"

    def __init__(
        self,
        source_detection: Union[Dict[str, Any], SourceDetection],
        edge: float = 0.1,
        bias: float = 0.0,
        saturated: float = 0.7,
        **kwargs: Any,
    ):
        """Create new exp time estimator from single star.

        Args:
            source_detection: Source detection to use.
            edge: Fraction of image to ignore at each border.
            bias: Bias level of image.
            saturated: Fraction of saturation that is used as brightness limit.
        """
        ExpTimeEstimator.__init__(self, **kwargs)

        self._source_detection = source_detection
        self._edge = edge
        self._bias = bias
        self._saturated = saturated
        self.coordinates = (None, None)

    async def __call__(self, image: Image) -> Image:
        """Processes an image and stores new exposure time in exp_time attribute.

        Args:
            image: Image to process.

        Returns:
            Original image.
        """

        # get object
        source_detection = get_object(self._source_detection, SourceDetection)

        # do photometry and get copy of catalog
        catalog = (await source_detection(image)).catalog
        if catalog is None:
            log.info("No catalog found in image.")
            return image

        # sort catalog by peak flux
        catalog.sort("peak")

        # saturation level
        if "DET-SATU" in image.header and "DET-GAIN" in image.header:
            saturation = image.header["DET-SATU"] / image.header["DET-GAIN"]
        else:
            saturation = 50000

        # get max peak flux that we allow
        max_peak = saturation * self._saturated

        # filter out all stars that are saturated
        catalog = catalog[catalog["peak"] <= max_peak]

        # get brightest star, get its peak flux and store its coordinates
        star = catalog[0]
        peak = star["peak"]
        log.info("Found peak of %.2f at %.1fx%.1f.", star["peak"], star["x"], star["y"])
        self.coordinates = (star["x"], star["y"])

        # get exposure time of image
        exp_time = image.header["EXPTIME"]

        # calculate new exposure time and return it
        self.exp_time = exp_time / (peak - self._bias) * (max_peak - self._bias)
        return image


__all__ = ["StarExpTimeEstimator"]



from abc import ABCMeta

from .IMotion import IMotion


class IRoof(IMotion, metaclass=ABCMeta):
    """The module controls a roof."""

    __module__ = "pyobs.interfaces"
    pass


__all__ = ["IRoof"]



from abc import ABCMeta, abstractmethod
from typing import Any

from .interface import Interface


class IPointingSeries(Interface, metaclass=ABCMeta):
    """The module provides the interface for a device that initializes and finalizes a pointing series and adds points
    to it."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def start_pointing_series(self, **kwargs: Any) -> str:
        """Start a new pointing series.

        Returns:
            A unique ID or filename, by which the series can be identified.
        """
        ...

    @abstractmethod
    async def stop_pointing_series(self, **kwargs: Any) -> None:
        """Stop a pointing series."""
        ...

    @abstractmethod
    async def add_pointing_measure(self, **kwargs: Any) -> None:
        """Add a new measurement to the pointing series."""
        ...


__all__ = ["IPointingSeries"]



"""
TODO: write docs
"""
__title__ = "Sky flat pointings"

from .base import SkyFlatsBasePointing
from .static import SkyFlatsStaticPointing


__all__ = ["SkyFlatsStaticPointing", "SkyFlatsBasePointing"]



from abc import ABCMeta, abstractmethod
from typing import Any

from .IMotion import IMotion


class IRotation(IMotion, metaclass=ABCMeta):
    """The module controls a device that can rotate."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def set_rotation(self, angle: float, **kwargs: Any) -> None:
        """Sets the rotation angle to the given value in degrees."""
        ...

    @abstractmethod
    async def get_rotation(self) -> float:
        """Returns the current rotation angle."""
        ...

    @abstractmethod
    async def track(self, ra: float, dec: float, **kwargs: Any) -> None:
        """Tracks the position angle of a rotator for an alt-az telescope."""
        ...


__all__ = ["IRotation"]



from threading import Thread
import logging
from typing import Optional, Any, cast

log = logging.getLogger(__name__)


class ThreadWithReturnValue(Thread):
    def __init__(self, *args: Any, **kwargs: Any):
        Thread.__init__(self, *args, **kwargs)
        self._return: Optional[Any] = None
        self._exception: Optional[Exception] = None

    def run(self) -> None:
        if self._target is not None:
            try:
                # save return valie
                self._return = self._target(*self._args, **self._kwargs)
            except Exception as e:
                # save exception, if one was caught
                self._exception = e

    def join(self, timeout: Optional[float] = None) -> Any:
        if timeout is None:
            log.warning("Joining thread with timeout of 0s. Is this correct?")
        # join thread
        Thread.join(self, timeout=timeout)
        # raise exception, if one was raised
        if self._exception is not None:
            raise self._exception
        # otherwise return value
        return self._return


__all__ = ["ThreadWithReturnValue"]



from abc import ABCMeta, abstractmethod
from typing import Tuple, Any

from .IAbortable import IAbortable


class IFlatField(IAbortable, metaclass=ABCMeta):
    """The module performs flat-fielding."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def flat_field(self, count: int = 20, **kwargs: Any) -> Tuple[int, float]:
        """Do a series of flat fields.

        Args:
            count: Number of images to take

        Returns:
            Number of images actually taken and total exposure time in ms
        """
        ...


__all__ = ["IFlatField"]



import logging
from typing import List, Dict, Optional, Any
from astroplan import (
    TimeConstraint,
    AirmassConstraint,
    ObservingBlock,
    FixedTarget,
    MoonSeparationConstraint,
    MoonIlluminationConstraint,
    AtNightConstraint,
)
from astropy.coordinates import SkyCoord
import astropy.units as u

from pyobs.utils.time import Time
from pyobs.robotic.taskarchive import TaskArchive
from .portal import Portal
from .task import LcoTask

log = logging.getLogger(__name__)


class LcoTaskArchive(TaskArchive):
    """Scheduler for using the LCO portal"""

    def __init__(
        self,
        url: str,
        token: str,
        instrument_type: str,
        **kwargs: Any,
    ):
        """Creates a new LCO scheduler.

        Args:
            url: URL to portal
            token: Authorization token for portal
            instrument_type: Type of instrument to use.
            scripts: External scripts
        """
        TaskArchive.__init__(self, **kwargs)

        # portal
        self._portal = Portal(url, token)

        # store stuff
        self._url = url
        self._token = token
        self._instrument_type = instrument_type

        # buffers in case of errors
        self._last_changed: Optional[Time] = None

        # task list
        self._tasks: Dict[str, LcoTask] = {}

    async def last_changed(self) -> Optional[Time]:
        """Returns time when last time any blocks changed."""

        # try to update time
        try:
            # get data
            self._last_changed = await self._portal.last_changed()

        finally:
            # even in case of errors, return last time
            return self._last_changed

    async def get_schedulable_blocks(self) -> List[ObservingBlock]:
        """Returns list of schedulable blocks.

        Returns:
            List of schedulable blocks
        """

        # get data
        schedulable = await self._portal.schedulable_requests()

        # get proposal priorities
        proposals = await self._portal.proposals()
        tac_priorities = {p["id"]: p["tac_priority"] for p in proposals}

        # loop all request groups
        blocks = []
        for group in schedulable:
            # get base priority, which is tac_priority * ipp_value
            proposal = group["proposal"]
            if proposal not in tac_priorities:
                log.error('Could not find proposal "%s".', proposal)
                continue
            base_priority = group["ipp_value"] * tac_priorities[proposal]

            # loop all requests in group
            for req in group["requests"]:
                # still pending?
                if req["state"] != "PENDING":
                    continue

                # duration
                duration = req["duration"] * u.second

                # time constraints
                time_constraints = [TimeConstraint(Time(wnd["start"]), Time(wnd["end"])) for wnd in req["windows"]]

                # loop configs
                for cfg in req["configurations"]:
                    # get instrument and check, whether we schedule it
                    instrument = cfg["instrument_type"]
                    if instrument.lower() != self._instrument_type.lower():
                        continue

                    # target
                    t = cfg["target"]
                    target = SkyCoord(t["ra"] * u.deg, t["dec"] * u.deg, frame=t["type"].lower())

                    # constraints
                    c = cfg["constraints"]
                    constraints = []
                    if "max_airmass" in c and c["max_airmass"] is not None:
                        constraints.append(AirmassConstraint(max=c["max_airmass"], boolean_constraint=False))
                    if "min_lunar_distance" in c and c["min_lunar_distance"] is not None:
                        constraints.append(MoonSeparationConstraint(min=c["min_lunar_distance"] * u.deg))
                    if "max_lunar_phase" in c and c["max_lunar_phase"] is not None:
                        constraints.append(MoonIlluminationConstraint(max=c["max_lunar_phase"]))
                        # if max lunar phase <= 0.4 (which would be DARK), we also enforce the sun to be <-18 degrees
                        if c["max_lunar_phase"] <= 0.4:
                            constraints.append(AtNightConstraint.twilight_astronomical())

                    # priority is base_priority times duration in minutes
                    # priority = base_priority * duration.value / 60.
                    priority = base_priority

                    # create block
                    block = ObservingBlock(
                        FixedTarget(target, name=req["id"]),
                        duration,
                        priority,
                        constraints=[*constraints, *time_constraints],
                        configuration={"request": req},
                    )
                    blocks.append(block)

        # return blocks
        return blocks


__all__ = ["LcoTaskArchive"]



"""
Misc
----
"""

from .addmask import AddMask
from .calibration import Calibration
from .createfilename import CreateFilename
from .removebackground import RemoveBackground
from .softbin import SoftBin
from .broadcast import Broadcast



import asyncio
import logging
import time

import numpy as np
from typing import Union, Optional, Any, Tuple, cast, Dict, List

from pyobs.interfaces import (
    IBinning,
    IWindow,
    IExposureTime,
    IRoof,
    IAutoGuiding,
    ITelescope,
    IAcquisition,
    ICamera,
    IFilters,
    IImageType,
)
from pyobs.robotic.scripts import Script
from pyobs.utils.enums import ImageType
from pyobs.utils.logger import DuplicateFilter
from pyobs.utils.parallel import Future
from pyobs.robotic import TaskRunner, TaskSchedule, TaskArchive

log = logging.getLogger(__name__)

# logger for logging name of task
cannot_run_logger = logging.getLogger(__name__ + ":cannot_run")
cannot_run_logger.addFilter(DuplicateFilter())


class LcoDefaultScript(Script):
    """Default script for LCO configs."""

    def __init__(
        self,
        camera: Union[str, ICamera],
        roof: Optional[Union[str, IRoof]] = None,
        telescope: Optional[Union[str, ITelescope]] = None,
        filters: Optional[Union[str, IFilters]] = None,
        autoguider: Optional[Union[str, IAutoGuiding]] = None,
        acquisition: Optional[Union[str, IAcquisition]] = None,
        **kwargs: Any,
    ):
        """Initialize a new LCO default script.

        Args:
            roof: Roof to use
            telescope: Telescope to use
            camera: Camera to use
            filters: Filter wheel to use
            autoguider: Autoguider to use
            acquisition: Acquisition to use
        """
        Script.__init__(self, **kwargs)

        # store
        self.roof = roof
        self.telescope = telescope
        self.camera = camera
        self.filters = filters
        self.autoguider = autoguider
        self.acquisition = acquisition

        # get image type
        self.image_type = ImageType.OBJECT
        if self.configuration["type"] == "BIAS":
            self.image_type = ImageType.BIAS
        elif self.configuration["type"] == "DARK":
            self.image_type = ImageType.DARK

    async def _get_proxies(
        self,
    ) -> Tuple[
        Optional[IRoof],
        Optional[ITelescope],
        Optional[ICamera],
        Optional[IFilters],
        Optional[IAutoGuiding],
        Optional[IAcquisition],
    ]:
        """Get proxies for running the task

        Returns:
            Proxies for roof, telescope, camera and filter wheel

        Raises:
            ValueError: If could not get proxies for all modules
        """
        roof = await self.comm.safe_proxy(self.roof, IRoof)
        telescope = await self.comm.safe_proxy(self.telescope, ITelescope)
        camera = await self.comm.safe_proxy(self.camera, ICamera)
        filters = await self.comm.safe_proxy(self.filters, IFilters)
        autoguider = await self.comm.safe_proxy(self.autoguider, IAutoGuiding)
        acquisition = await self.comm.safe_proxy(self.acquisition, IAcquisition)
        return roof, telescope, camera, filters, autoguider, acquisition

    async def can_run(self) -> bool:
        """Whether this config can currently run.

        Returns:
            True, if script can run now
        """

        # get proxies
        roof, telescope, camera, filters, autoguider, acquisition = await self._get_proxies()

        # need camera
        if camera is None:
            cannot_run_logger.info("Cannot run task, no camera found.")
            return False

        # for OBJECT exposure we need more
        if self.image_type == ImageType.OBJECT:
            # we need an open roof and a working telescope
            if roof is None or not await roof.is_ready():
                cannot_run_logger.warning("Cannot run task, no roof found or roof not ready.")
                return False
            if telescope is None or not await telescope.is_ready():
                cannot_run_logger.warning("Cannot run task, no telescope found or telescope not ready.")
                return False

            # we probably need filters and autoguider/acquisition
            if filters is None:
                cannot_run_logger.warning("Cannot run task, No filter module found.")
                return False

            # acquisition?
            if (
                "acquisition_config" in self.configuration
                and "mode" in self.configuration["acquisition_config"]
                and self.configuration["acquisition_config"]["mode"] == "ON"
                and acquisition is None
            ):
                cannot_run_logger.warning("Cannot run task, no acquisition found.")
                return False

            # guiding?
            if (
                "guiding_config" in self.configuration
                and "mode" in self.configuration["guiding_config"]
                and self.configuration["guiding_config"]["mode"] == "ON"
                and autoguider is None
            ):
                cannot_run_logger.warning("Cannot run task, no auto guider found.")
                return False

        # seems alright
        return True

    async def run(
        self,
        task_runner: TaskRunner,
        task_schedule: Optional[TaskSchedule] = None,
        task_archive: Optional[TaskArchive] = None,
    ) -> None:
        """Run script.

        Raises:
            InterruptedError: If interrupted
        """

        # get proxies
        roof, telescope, camera, filters, autoguider, acquisition = await self._get_proxies()

        # got a target?
        target = self.configuration["target"]
        track: Union[Future, asyncio.Task[Any]] = Future(empty=True)
        if self.image_type == ImageType.OBJECT:
            if telescope is None:
                raise ValueError("No telescope given.")
            log.info("Moving to target %s...", target["name"])
            track = asyncio.create_task(telescope.move_radec(target["ra"], target["dec"]))

        # acquisition?
        if (
            "acquisition_config" in self.configuration
            and "mode" in self.configuration["acquisition_config"]
            and self.configuration["acquisition_config"]["mode"] == "ON"
        ):
            # wait for track
            await track

            # do acquisition
            if acquisition is None:
                raise ValueError("No acquisition given.")
            log.info("Performing acquisition...")
            await acquisition.acquire_target()

        # guiding?
        if (
            "guiding_config" in self.configuration
            and "mode" in self.configuration["guiding_config"]
            and self.configuration["guiding_config"]["mode"] == "ON"
        ):
            if autoguider is None:
                raise ValueError("No autoguider given.")
            log.info("Starting auto-guiding...")
            await autoguider.start()

        # total (exposure) time done in this config
        self.exptime_done = 0.0

        # task archive must be LCO
        from pyobs.robotic.lco import LcoTaskSchedule

        if not isinstance(task_schedule, LcoTaskSchedule):
            raise ValueError("Task schedule is not for LCO observation portal.")

        # get instrument info
        instrument_type = self.configuration["instrument_type"].lower()
        instrument = task_schedule.instruments[instrument_type]

        # setting repeat duration depending on config type
        repeat_duration = None
        if self.configuration["type"] == "REPEAT_EXPOSE":
            if "repeat_duration" in self.configuration:
                repeat_duration = self.configuration["repeat_duration"]
                log.info("Repeating all instrument configurations for %d seconds.", repeat_duration)
            else:
                log.error("Type is REPEAT_EXPOSE, but no repeat_duration was set.")

        # config iterations
        config_finished = False
        ic_durations = []
        image_no = 1
        while not config_finished:
            # ic start time
            ic_start_time = time.time()

            # loop instrument configs
            for ic in self.configuration["instrument_configs"]:
                # get readout mode
                for readout_mode in instrument["modes"]["readout"]["modes"]:
                    if readout_mode["code"] == ic["mode"]:
                        break
                else:
                    # could not find readout mode
                    raise ValueError("Could not find readout mode %s." % ic["mode"])
                log.info('Using readout mode "%s"...' % readout_mode["name"])

                # set filter
                set_filter: Union[Future, asyncio.Task[Any]] = Future(empty=True)
                if "optical_elements" in ic and "filter" in ic["optical_elements"] and filters is not None:
                    log.info("Setting filter to %s...", ic["optical_elements"]["filter"])
                    set_filter = asyncio.create_task(filters.set_filter(ic["optical_elements"]["filter"]))

                # wait for tracking and filter
                await Future.wait_all([track, set_filter])

                # set binning and window
                if isinstance(camera, IBinning):
                    bin_x = readout_mode["validation_schema"]["bin_x"]["default"]
                    bin_y = readout_mode["validation_schema"]["bin_y"]["default"]
                    log.info("Set binning to %dx%d...", bin_x, bin_y)
                    await camera.set_binning(bin_x, bin_y)
                if isinstance(camera, IWindow):
                    full_frame = await camera.get_full_frame()
                    await camera.set_window(*full_frame)

                # loop images
                for exp in range(ic["exposure_count"]):
                    # prepare log entry
                    # add total image number and number within IC
                    msg = f"Exposing {self.configuration['type']} image #{image_no} ({exp + 1}/{ic['exposure_count']})"

                    # set exposure time
                    if isinstance(camera, IExposureTime):
                        await camera.set_exposure_time(ic["exposure_time"])
                        msg += f" for {ic['exposure_time']:2f}s"

                    # set image type
                    if isinstance(camera, IImageType):
                        await camera.set_image_type(self.image_type)

                    # log it
                    log.info(f"{msg}...")

                    # grab image
                    await cast(ICamera, camera).grab_image()
                    self.exptime_done += ic["exposure_time"]
                    image_no += 1

            # store duration for all ICs
            ic_durations.append(time.time() - ic_start_time)

            # need repeat?
            if repeat_duration is None:
                # if there is no repeat duration, we're finished
                config_finished = True

            else:
                # get average IC duration
                avg_ic_duration = np.mean(ic_durations)

                # can we do another one, i.e. is done time plus average time larger than repeat_duration?
                if sum(ic_durations) + avg_ic_duration > repeat_duration:
                    # doesn't seem so
                    config_finished = True

        # stop auto guiding
        if (
            "guiding_config" in self.configuration
            and "mode" in self.configuration["guiding_config"]
            and self.configuration["guiding_config"]["mode"] == "ON"
            and autoguider is not None
        ):
            log.info("Stopping auto-guiding...")
            await autoguider.stop()

        # finally, stop telescope
        if self.image_type == ImageType.OBJECT:
            log.info("Stopping telescope...")
            await cast(ITelescope, telescope).stop_motion()

    def get_fits_headers(self, namespaces: Optional[List[str]] = None) -> Dict[str, Any]:
        """Returns FITS header for the current status of this module.

        Args:
            namespaces: If given, only return FITS headers for the given namespaces.

        Returns:
            Dictionary containing FITS headers.
        """

        # init header
        hdr = {}

        # which image type?
        if self.image_type == ImageType.OBJECT:
            # add object name
            hdr["OBJECT"] = self.configuration["target"]["name"], "Name of target"

        # return
        return hdr


__all__ = ["LcoDefaultScript"]



import logging
from typing import Any, Optional

from pyobs.modules import Module
from pyobs.events import LogEvent, Event
from pyobs.utils.time import Time

log = logging.getLogger(__name__)


class FluentLogger(Module):
    """Log to fluentd server."""

    __module__ = "pyobs.modules.utils"

    def __init__(self, hostname: str, port: int, *args: Any, **kwargs: Any):
        """Initialize a new logger.

        Args:
            hostname: Hostname of server.
            port: Port of server.

        """
        from fluent import sender

        Module.__init__(self, **kwargs)

        # store
        self._hostname = hostname
        self._port = port
        self._fluent: Optional[sender.FluentSender] = None

    async def open(self) -> None:
        """Open module."""
        from fluent import sender

        await Module.open(self)

        # get handler
        self._fluent = sender.FluentSender("pyobs", host=self._hostname, port=self._port)

        # listen to log events
        await self.comm.register_event(LogEvent, self._process_log_entry)

    async def _process_log_entry(self, event: Event, sender: str) -> bool:
        """Process a new log entry.

        Args:
            event: The log event.
            sender: Name of sender.
        """

        # check
        if self._fluent is None:
            raise ValueError("Module not opened.")
        if not isinstance(event, LogEvent):
            raise ValueError("Wrong event type.")

        # get time
        time = Time(event.time).unix

        # send it
        self._fluent.emit_with_time(sender, time, event.data)
        return True


__all__ = ["FluentLogger"]



from typing import Optional, Any
from typing_extensions import TypedDict

from pyobs.events.event import Event


DataType = TypedDict("DataType", {"name": str, "id": Any})


class TaskFinishedEvent(Event):
    """Event to be sent when a task has finished."""

    __module__ = "pyobs.events"

    def __init__(self, name: str, id: Any, **kwargs: Any):
        """Initializes a new task finished event.

        Args:
            name: Name of task that just finished
            id: Unique identifier for task
        """
        Event.__init__(self)
        self.data: DataType = {"name": name, "id": id}

    @property
    def name(self) -> str:
        return self.data["name"]

    @property
    def id(self) -> Any:
        return self.data["id"]


__all__ = ["TaskFinishedEvent"]



import logging
from typing import Union, Dict, Any

from pyobs.interfaces import IRunnable, ITelescope
from pyobs.modules import Module
from pyobs.object import get_object
from pyobs.modules import timeout
from pyobs.utils.skyflats.pointing.base import SkyFlatsBasePointing

log = logging.getLogger(__name__)


class FlatFieldPointing(Module, IRunnable):
    """Module for pointing a telescope."""

    __module__ = "pyobs.modules.flatfield"

    def __init__(
        self, telescope: Union[str, ITelescope], pointing: Union[Dict[str, Any], SkyFlatsBasePointing], **kwargs: Any
    ):
        """Initialize a new flat field pointing.

        Args:
            telescope: Telescope to point
            pointing: Pointing for calculating coordinates.
        """
        Module.__init__(self, **kwargs)

        # store telescope and pointing
        self._telescope = telescope
        self._pointing = pointing

    @timeout(60)
    async def run(self, **kwargs: Any) -> None:
        """Move telescope to pointing."""

        # get telescope
        log.info("Getting proxy for telescope...")
        telescope = await self.proxy(self._telescope, ITelescope)

        # pointing
        pointing = get_object(self._pointing, SkyFlatsBasePointing, observer=self.observer)

        # point
        await pointing(telescope)
        log.info("Finished pointing telescope.")

    def abort(self, **kwargs: Any) -> None:
        """Abort current actions."""
        pass


__all__ = ["FlatFieldPointing"]



import logging
import os.path
from typing import Union, Dict, Tuple, Optional, Any

from pyobs.object import get_object
from pyobs.utils.time import Time
from pyobs.utils.fits import FilenameFormatter
from pyobs.images import Image
from pyobs.utils.archive import Archive
from pyobs.utils.enums import ImageType
from .pipeline import Pipeline

log = logging.getLogger(__name__)


FILENAME = "{SITEID}{TELID}-{INSTRUME}-{DAY-OBS|date:}-{IMAGETYP}-{XBINNING}x{YBINNING}{FILTER|filter}.fits"


class Night:
    def __init__(
        self,
        archive: Union[Dict[str, Any], Archive],
        pipeline: Union[Dict[str, Any], Pipeline],
        worker_procs: int = 4,
        filenames_calib: str = FILENAME,
        min_flats: int = 10,
        store_local: Optional[str] = None,
        create_calibs: bool = True,
        calib_science: bool = True,
        **kwargs: Any,
    ):
        """Creates a Night object for reducing a given night.

        Args:
            archive: Archive to fetch images from and write results to.
            pipeline: Science pipeline.
            worker_procs: Number of worker processes.
            filenames_calib: Filename pattern for master calibration files.
            min_flats: Minimum number of raw frames to create flat field.
            store_local: If True, files are stored in given local directory instead of uploaded to archive.
            create_calibs: If False, no calibration files are created for night.
            calib_science: If False, no science frames are calibrated.
        """

        # get archive and science pipeline
        self._archive = get_object(archive, Archive)
        self._pipeline = get_object(pipeline, Pipeline, archive=archive)

        # stuff
        self._worker_processes = worker_procs
        self._min_flats = min_flats
        self._store_local = store_local
        self._create_calibs = create_calibs
        self._calib_science = calib_science

        # cache for master calibration frames
        self._master_frames: Dict[Tuple[ImageType, str, str, Optional[str]], Image] = {}

        # default filename patterns
        self._fmt_calib = FilenameFormatter(filenames_calib)

    async def _find_master(
        self,
        night: str,
        image_type: ImageType,
        instrument: str,
        binning: str,
        filter_name: Optional[str] = None,
        max_days: float = 30.0,
    ) -> Optional[Image]:
        """Find master calibration frame for given parameters using a cache.

        Args:
            image_type: image type.
            instrument: Instrument name.
            binning: Binning.
            filter_name: Name of filter.
            max_days: Maximum number of days from DATE-OBS to find frames.

        Returns:
            Image or None
        """

        # is in cache?
        if (image_type, instrument, binning, filter_name) in self._master_frames:
            return self._master_frames[image_type, instrument, binning, filter_name]

        # try to download one
        midnight = Time(night + " 23:59:59")
        image = await self._pipeline.find_master(
            self._archive, image_type, midnight, instrument, binning, filter_name, max_days=max_days
        )
        if image is not None:
            # store and return it
            self._master_frames[image_type, instrument, binning, filter_name] = image
            return image
        else:
            # still nothing
            return None

    async def _create_master_calib(
        self, night: str, instrument: str, image_type: ImageType, binning: str, filter_name: Optional[str] = None
    ) -> Optional[Image]:
        # get frames
        infos = await self._archive.list_frames(
            night=night,
            image_type=image_type,
            filter_name=filter_name,
            instrument=instrument,
            binning=binning,
            rlevel=0,
        )

        # log it
        fltr = "" if filter_name is None else " in " + filter_name
        log.info("Found %d %s %s frames%s from instrument %s.", len(infos), binning, image_type.value, fltr, instrument)

        # if too few, we're finished
        if len(infos) < 3:
            if len(infos) > 0:
                log.warning("Too few (%d) frames found, skipping...", len(infos))
            return None

        # download frames
        images = await self._archive.download_frames(infos)
        if len(images) < 3:
            log.warning("Too few (%d) frames found, skipping...", len(infos))

        # create master
        calib: Optional[Image] = None
        if image_type == ImageType.BIAS:
            # BIAS are easy, just combine
            calib = await self._pipeline.create_master_bias(images)
            if calib is None:
                log.warning("Could not create master bias.")
                return None

            # store in cache
            self._master_frames[ImageType.BIAS, instrument, binning, None] = calib

        elif image_type == ImageType.DARK:
            # for DARKs, we first need a BIAS
            bias = await self._find_master(night, ImageType.BIAS, instrument, binning, None)
            if bias is None:
                log.error("Could not find BIAS frame, skipping...")
                return None

            # combine
            calib = await self._pipeline.create_master_dark(images, bias=bias)
            if calib is None:
                log.warning("Could not create master dark.")
                return None

            # store in cache
            self._master_frames[ImageType.DARK, instrument, binning, None] = calib

        elif image_type == ImageType.SKYFLAT:
            # got enough frames?
            if len(images) < self._min_flats:
                log.warning("Not enough flat fields found for combining.")
                return None

            # for SKYFLATs, we first need a BIAS
            bias = await self._find_master(night, ImageType.BIAS, instrument, binning, None)
            if bias is None:
                log.error("Could not find BIAS frame, skipping...")
                return None

            # combine
            calib = await self._pipeline.create_master_flat(images, bias=bias)
            if calib is None:
                log.warning("Could not create master flat.")
                return None

            # store in cache
            self._master_frames[ImageType.SKYFLAT, instrument, binning, filter_name] = calib

        else:
            raise ValueError("Invalid image type")

        # filename
        calib.format_filename(self._fmt_calib)

        # save/upload
        if self._store_local:
            path = os.path.join(self._store_local, calib.header["FNAME"])
            log.info("Storing master calibration frame as %s...", path)
            calib.writeto(path, overwrite=True)
        else:
            log.info("Uploading master calibration frame as %s...", calib.header["FNAME"])
            await self._archive.upload_frames([calib])

        # finished
        return calib

    async def _calib_data(self, night: str, instrument: str, binning: str, filter_name: str) -> None:
        # get all frames
        infos = await self._archive.list_frames(
            night=night,
            instrument=instrument,
            image_type=ImageType.OBJECT,
            binning=binning,
            filter_name=filter_name,
            rlevel=0,
        )
        total = len(infos)
        if total == 0:
            return
        log.info("Calibrating %d OBJECT frames...", total)

        # run all science frames
        for i, info in enumerate(infos, 1):
            log.info("(%d/%d) Calibrating file %s...", i, total, info.filename)

            try:
                # download frame
                images = await self._archive.download_frames([info])
                image = images[0]

                # calibrate
                calibrated = await self._pipeline.calibrate(image)

                # save/upload
                if self._store_local:
                    path = os.path.join(self._store_local, calibrated.header["FNAME"])
                    log.info("(%d/%d) Storing calibrated images as %s...", i, total, path)
                    calibrated.writeto(path, overwrite=True)
                else:
                    log.info("(%d/%d) Uploading calibrated images as %s...", i, total, calibrated.header["FNAME"])
                    await self._archive.upload_frames([calibrated])

            except Exception:
                log.exception("(%d/%d) Error processing image %s.", i, total, info.filename)

    async def __call__(self, site: str, night: str) -> None:
        """Reduces all data im this night."""

        # get options
        log.info("Retrieving configurations for site %s at night %s...", site, night)
        options = await self._archive.list_options(night=night, site=site)
        log.info(
            "Got data for %d instruments, %d binnings, and %d filters.",
            len(options["instruments"]),
            len(options["binnings"]),
            len(options["filters"]),
        )

        # loop instruments
        for instrument in options["instruments"]:
            log.info("Reducing data for instrument %s...", instrument)

            # loop binnings
            for binning in options["binnings"]:
                # create bias and dark
                if self._create_calibs:
                    await self._create_master_calib(night, instrument, ImageType.BIAS, binning)
                    await self._create_master_calib(night, instrument, ImageType.DARK, binning)

                # loop filters
                for filter_name in options["filters"]:
                    # create flat
                    if self._create_calibs:
                        await self._create_master_calib(night, instrument, ImageType.SKYFLAT, binning, filter_name)

                    # calibrate science data
                    if self._calib_science:
                        await self._calib_data(night, instrument, binning, filter_name)


__all__ = ["Night"]



import logging
from typing import Tuple, Any

from pyobs.images import Image
from pyobs.images.meta import PixelOffsets
from .offsets import Offsets

log = logging.getLogger(__name__)


class BrightestStarOffsets(Offsets):
    """Calculates offsets from the center of the image to the brightest star."""

    __module__ = "pyobs.images.processors.offsets"

    def __init__(self, center: Tuple[str, str] = ("CRPIX1", "CRPIX2"), **kwargs: Any):
        """Initializes a new auto guiding system."""
        Offsets.__init__(self, **kwargs)

        # init
        self._center = center

    async def __call__(self, image: Image) -> Image:
        """Processes an image and sets x/y pixel offset to reference in offset attribute.

        Args:
            image: Image to process.

        Returns:
            Original image.

        Raises:
            ValueError: If offset could not be found.
        """

        # get catalog and sort by flux
        cat = image.catalog
        if cat is None or len(cat) < 1:
            log.warning("No catalog found in image.")
            return image
        cat.sort("flux", reverse=True)

        # get first X/Y coordinates
        x, y = cat["x"][0], cat["y"][0]

        # get center
        center_x, center_y = image.header[self._center[0]], image.header[self._center[1]]

        # calculate offset
        dx, dy = center_x - x, center_y - y

        # set it and return image
        image.set_meta(PixelOffsets(dx, dy))
        return image


__all__ = ["BrightestStarOffsets"]



from typing import List, Tuple

import numpy as np
from scipy.optimize import curve_fit


def fit_hyperbola(x_arr: List[float], y_arr: List[float], y_err: List[float]) -> Tuple[float, float]:
    """Fit a hyperbola

    Args:
        x_arr: X data
        y_arr: Y data
        y_err: Y errors

    Returns:
        Minimum of hyperbola and its uncertainty
    """

    # initial guess
    ic = np.argmin(y_arr)
    ix = np.argmax(y_arr)
    b = y_arr[ic]
    c = x_arr[ic]
    x = x_arr[ix]
    slope = np.abs((y_arr[ic] - y_arr[ix]) / (c - x))
    a = b / slope

    # init
    p0 = [a, b, c]

    # fit
    coeffs, cov = curve_fit(
        lambda xx, aa, bb, cc: bb * np.sqrt((xx - cc) ** 2 / aa ** 2 + 1.0), x_arr, y_arr, sigma=y_err, p0=p0
    )

    # return result
    return coeffs[2], cov[2][2]



import asyncio
import logging
from typing import Optional, Any, Dict, TYPE_CHECKING, cast
from py_expression_eval import Parser
import pandas as pd
import numpy as np
import numpy.typing as npt

if TYPE_CHECKING:
    import lmfit

from pyobs.interfaces import IFocuser, IFilters, IWeather, ITemperatures
from pyobs.modules import Module
from pyobs.modules import timeout
from pyobs.interfaces import IFocusModel
from pyobs.events import FocusFoundEvent, FilterChangedEvent, Event
from pyobs.utils.enums import WeatherSensors
from pyobs.utils.publisher import CsvPublisher
from pyobs.utils.time import Time

log = logging.getLogger(__name__)


class FocusModel(Module, IFocusModel):
    """A focus model that is automatically applied to an IFocuser.

    If, e.g., the model is defined as:

        model: -0.043807*T1 - 0.031798*T2 + 0.062042*temp + 41.694895

    Then "temp" is taken from the weather module automatically. The other temperatures must be defined, e.g., as:

        temperatures:
            T1:
                module: telescope
                sensor: T1
            T2:
                module: telescope
                sensor: T2

    In this case, the method get_temperatures() is called on the module "telescope" and the values T1 and T2 are
    taken for the model.

    Alternatively, the coefficients can be defined as symbols:

        model: a*T1 + b*T2 + c*temp + d

    For this to work, initial values must be specified separately:

        coefficients:
            a: -0.043807
            b: -0.031798
            c: 0.062042
            d: 41.694895

    Only this way it is possible to automatically re-calculate the model.
    """

    __module__ = "pyobs.modules.focus"

    def __init__(
        self,
        focuser: Optional[str] = None,
        weather: Optional[str] = None,
        interval: int = 300,
        temperatures: Optional[Dict[str, Dict[str, float]]] = None,
        model: Optional[str] = None,
        coefficients: Optional[Dict[str, float]] = None,
        update: bool = False,
        log_file: Optional[str] = None,
        min_measurements: int = 10,
        enabled: bool = True,
        temp_sensor: str = "average.temp",
        default_filter: Optional[str] = None,
        filter_offsets: Optional[Dict[str, float]] = None,
        filter_wheel: Optional[str] = None,
        **kwargs: Any,
    ):
        """Initialize a focus model.

        Args:
            focuser: Name of focuser.
            weather: Name of weather station.
            interval: Interval for setting focus or None, if no regular setting of focus is required.
            model: Focus model to use.
            coefficients: Coefficients in model, mainly used when updating it.
            update: Whether to update the model on new focus values.
            log_file: Path to file containing all focus measurements.
            min_measurements: Minimum number of measurements to update model.
            enabled: If False, no focus is set.
            temp_sensor: Name of sensor at weather station to provide ambient temperature.
            default_filter: Name of default filter. If None, filters are ignored.
            filter_offsets: Offsets for different filters. If None, they are not modeled.
            filter_wheel: Name of filter wheel module to use for fetching filter before setting focus.
        """
        Module.__init__(self, **kwargs)

        # check import
        import lmfit

        log.info(f"Found lmfit {lmfit.__version__}.")

        # add thread func
        if interval is not None and interval > 0:
            self.add_background_task(self._update)

        # store
        self._focuser = focuser
        self._weather = weather
        self._interval = interval
        self._temperatures: Dict[str, Dict[str, float]] = {} if temperatures is None else temperatures
        self._focuser_ready = True
        self._coefficients = {} if coefficients is None else coefficients
        self._update_model = update
        self._min_measurements = min_measurements
        self._enabled = enabled
        self._temp_station, sensor = temp_sensor.split(".")
        self._temp_sensor = WeatherSensors(sensor)
        self._default_filter = default_filter
        self._filter_offsets = filter_offsets
        self._filter_wheel = filter_wheel
        log.info("Going to fetch temperature from sensor %s at station %s.", self._temp_sensor, self._temp_station)

        # model
        parser = Parser()
        log.info("Parsing model: %s", model)
        self._temp_model = parser.parse(model)

        # coefficients
        if self._coefficients is not None and len(self._coefficients) > 0:
            log.info("Found coefficients: %s", ", ".join(["%s=%.3f" % (k, v) for k, v in self._coefficients.items()]))

        # variables
        variables = self._temp_model.variables()
        for c in self._coefficients.keys():
            variables.remove(c)
        log.info("Found variables: %s", ", ".join(variables))

        # init log file
        self._publisher = None if log_file is None else CsvPublisher(log_file)

    async def open(self) -> None:
        """Open module."""
        await Module.open(self)

        # subscribe to events
        await self.comm.register_event(FocusFoundEvent, self._on_focus_found)
        if self._filter_offsets is not None and self._filter_wheel is not None:
            await self.comm.register_event(FilterChangedEvent, self._on_filter_changed)

        # update model now
        if self._update_model:
            await self._calc_focus_model()

    async def _update(self) -> None:
        # wait a little
        await asyncio.sleep(1)

        # run until closed
        while True:
            # if not enabled, just sleep a little
            if not self._enabled:
                await asyncio.sleep(1)
                continue

            # get focuser
            try:
                focuser = await self.proxy(self._focuser, IFocuser)
            except ValueError:
                log.warning("Could not connect to focuser.")
                await asyncio.sleep(10)
                continue

            # is focuser ready?
            status = await focuser.is_ready()
            if status is False:
                # log
                if self._focuser_ready:
                    log.info("Focuser not ready, waiting for it...")
                    self._focuser_ready = False

                # sleep a little and continue
                await asyncio.sleep(10)
                continue

            # came from not ready state?
            if not self._focuser_ready:
                log.info("Focuser ready now, starting focus tracking...")
                self._focuser_ready = True

            # now set focus
            try:
                # set optimal focus
                await self.set_optimal_focus()
            except ValueError:
                # something went wrong, wait a little and continue
                await asyncio.sleep(10)
                continue

            # sleep interval
            log.info("Going to sleep for %d seconds...", self._interval)
            await asyncio.sleep(self._interval)

    async def _get_optimal_focus(self, filter_name: Optional[str] = None, **kwargs: Any) -> float:
        """Returns the optimal focus.

        Args:
            filter_name: If given, use this filter name instead of fetching one.

        Returns:
            Optimum focus calculated from model.

        Raises:
            ValueError: If anything went wrong.
        """

        # get values for variables
        values = await self._get_values()

        # evaluate model
        log.info("Evaluating model...")
        focus = self._temp_model.evaluate({**values, **self._coefficients})

        # focus offset?
        if self._filter_offsets is not None and self._filter_wheel is not None:
            try:
                # need a filter name?
                if filter_name is None:
                    # get proxy
                    wheel = await self.proxy(self._filter_wheel, IFilters)

                    # get filter
                    filter_name = await wheel.get_filter()

                # add offset
                offset = self._filter_offsets[filter_name]
                log.info("Adding filter offset of %.2f for filter %s...", offset, filter_name)
                focus += offset

            except (ValueError, KeyError):
                log.error("Could not determine filter offset.")

        # set focus
        log.info("Found optimal focus of %.4f.", focus)
        return float(focus)

    async def get_optimal_focus(self, **kwargs: Any) -> float:
        """Returns the optimal focus.

        Returns:
            Optimum focus calculated from model.

        Raises:
            ValueError: If anything went wrong.
        """
        return await self._get_optimal_focus()

    async def _get_values(self) -> Dict[str, Any]:
        """Retrieve all required values for the model.

        Returns:
            Dictionary containing all values required by the model.
        """

        # variables for model evaluation
        variables = {}

        # do we need a weather proxy?
        if "temp" in self._temp_model.variables():
            log.info("Fetching temperature from weather module...")

            # get weather proxy
            try:
                weather = await self.proxy(self._weather, IWeather)
            except ValueError:
                raise ValueError("Could not connect to weather module.")

            # get value
            time, val = await weather.get_sensor_value(self._temp_station, self._temp_sensor)
            if val is None:
                raise ValueError("Received invalid temperature from weather station.")

            # get temperature
            variables["temp"] = val
            log.info("Got temperature of %.2f.", variables["temp"])

        # loop other temperatures
        module_temps = {}
        for var, cfg in self._temperatures.items():
            # need to fetch temperatures from module?
            if cfg["module"] not in module_temps:
                log.info("Fetching temperatures from module %s...", cfg["module"])

                # get proxy
                proxy = await self.proxy(cfg["module"], ITemperatures)

                # get temperatures
                module_temps[cfg["module"]] = await proxy.get_temperatures()

                # log
                vals = ", ".join(["%s=%.2f" % (k, v) for k, v in module_temps[cfg["module"]].items()])
                log.info("Received temperatures: %s", vals)

            # store, what we need
            if cfg["sensor"] not in module_temps[cfg["module"]]:
                raise ValueError(
                    "Temperature for sensor %s not in data from module %s." % (cfg["sensor"], cfg["module"])
                )
            variables[var] = module_temps[cfg["module"]][cfg["sensor"]]

        # log
        vals = ", ".join(["%s=%.2f" % (k, v) for k, v in variables.items()])
        log.info("Found values for model: %s", vals)
        return variables

    async def _set_optimal_focus(self, filter_name: Optional[str] = None, **kwargs: Any) -> None:
        """Sets optimal focus.

        Args:
            filter_name: Name of filter to use.

        Raises:
            ValueError: If anything went wrong.
        """

        # get focuser
        focuser = await self.proxy(self._focuser, IFocuser)

        # get focus
        focus = await self._get_optimal_focus(filter_name=filter_name)

        # set it
        log.info("Setting optimal focus...")
        await focuser.set_focus(focus)
        log.info("Done.")

    @timeout(60)
    async def set_optimal_focus(self, **kwargs: Any) -> None:
        """Sets optimal focus.

        Raises:
            ValueError: If anything went wrong.
        """
        await self._set_optimal_focus()

    async def _on_focus_found(self, event: Event, sender: str) -> bool:
        """Receive FocusFoundEvent.

        Args:
            event: The event itself
            sender: The name of the sender.
        """
        if not isinstance(event, FocusFoundEvent):
            raise ValueError("Not a focus event.")
        log.info("Received new focus of %.4f +- %.4f.", event.focus, event.error)

        # collect values for model
        values = await self._get_values()

        # add focus and datetime
        values["focus"] = event.focus
        values["error"] = event.error
        values["datetime"] = Time.now().isot
        values["filter"] = event.filter_name

        # write log
        if self._publisher is not None:
            await self._publisher(**values)

        # finally, calculate new model
        log.info("Re-calculating model...")
        await self._calc_focus_model()

        # finished
        log.info("Done.")
        return True

    async def _calc_focus_model(self) -> None:
        """Calculate new focus model from saved entries."""
        import lmfit

        # no coefficients? no model...
        if not self._coefficients or self._publisher is None:
            return

        # only take clear filter images for now
        data = await self._publisher.data()

        # enough measurements?
        if len(data) < self._min_measurements:
            log.warning(
                "Not enough measurements found for re-calculating model (%d<%d).", len(data), self._min_measurements
            )
            return

        # build parameters
        params = lmfit.Parameters()
        for c in self._coefficients.keys():
            params.add(c, 0.0)

        # if we want to fit filter offsets, add them to params
        if self._filter_offsets is not None:
            # get unique list of filters and add them
            for f in data["filter"].unique():
                params.add("off_" + f, 0.0)

        # fit
        log.info("Fitting coefficients...")
        out = lmfit.minimize(self._residuals, params, args=(data,))
        if not hasattr(out, "params"):
            raise ValueError("No params returned from fit.")
        out_params = getattr(out, "params")

        # print results
        log.info("Found best coefficients:")
        for p in out_params:
            if not p.startswith("off_"):
                if out_params[p].stderr is not None:
                    log.info("  %-5s = %10.5f +- %8.5f", p, out_params[p].value, out_params[p].stderr)
                else:
                    log.info("  %-5s = %10.5f", p, out_params[p].value)
        if self._filter_offsets is not None:
            log.info("Found filter offsets:")
            for p in out_params:
                if p.startswith("off_"):
                    if out_params[p].stderr is not None:
                        log.info("  %-10s = %10.5f +- %8.5f", p[4:], out_params[p].value, out_params[p].stderr)
                    else:
                        log.info("  %-10s = %10.5f", p[4:], out_params[p].value)

        # evaluate model
        best_fit = self._model(out_params, data)
        real_focus = data["focus"]

        # calculate RMS and reduced chi square
        rms = np.sqrt(np.mean((best_fit - real_focus) ** 2))
        redchi = np.sum((best_fit - real_focus) ** 2) / out.nfree
        log.info("Reduced chi squared: %.3f, RMS: %.3f", redchi, rms)

        # store new coefficients and filter offsets
        if self._update_model:
            # just copy all?
            d = dict(out_params.valuesdict())
            if self._filter_offsets is None:
                self._coefficients = d
            else:
                # need to separate
                self._coefficients = {k: v for k, v in d.items() if not k.startswith("off_")}
                self._filter_offsets = {k[4:]: v for k, v in d.items() if k.startswith("off_")}

    def _model(self, x: "lmfit.Parameters", data: pd.DataFrame) -> npt.NDArray[float]:
        # calc model
        model = []
        for _, row in data.iterrows():
            # how do we fit?
            if self._default_filter is None:
                # just fit it
                mod = self._temp_model.evaluate({**x.valuesdict(), **row})

            else:
                # do we want to fit filter offsets?
                if self._filter_offsets:
                    # evaluate and add offset
                    mod = self._temp_model.evaluate({**x.valuesdict(), **row})
                    if row["filter"] != self._default_filter:
                        mod += x["off_" + row["filter"]]

                else:
                    # no filter offsets, so ignore this row if, if wrong filter
                    if row["filter"] == self._default_filter:
                        mod = self._temp_model.evaluate({**x.valuesdict(), **row})
                    else:
                        continue

            # add it
            model.append(mod)

        # finished
        return np.array(model)

    def _residuals(self, x: "lmfit.Parameters", data: pd.DataFrame) -> npt.NDArray[float]:
        """Fit method for model

        Args:
            x: Paramaters to evaluate.
            data: Full data set.

        Returns:

        """

        # get model and fitted focus/error
        model = self._model(x, data)
        focus = data["focus"]
        error = data["error"]

        # return residuals
        return cast(npt.NDArray[float], (focus - model) / error)

    async def _on_filter_changed(self, event: Event, sender: str) -> bool:
        """Receive FilterChangedEvent and set focus.

        Args:
            event: The event itself
            sender: The name of the sender.
        """

        # wrong sender?
        if sender != self._filter_wheel or not isinstance(event, FilterChangedEvent):
            return False

        # log and change
        try:
            log.info("Detected filter change to %s, adjusting focus...", event.filter)
            await self._set_optimal_focus(event.filter)
            return True
        except ValueError:
            log.error("Could not set focus.")
            return False


__all__ = ["FocusModel"]



from abc import ABCMeta, abstractmethod
from typing import Tuple, Any, Dict

from .IAbortable import IAbortable


class IAutoFocus(IAbortable, metaclass=ABCMeta):
    """The module can perform an auto-focus."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def auto_focus(self, count: int, step: float, exposure_time: float, **kwargs: Any) -> Tuple[float, float]:
        """Perform an auto-focus series.

        This method performs an auto-focus series with "count" images on each side of the initial guess and the given
        step size. With count=3, step=1 and guess=10, this takes images at the following focus values:
        7, 8, 9, 10, 11, 12, 13

        Args:
            count: Number of images to take on each side of the initial guess. Should be an odd number.
            step: Step size.
            exposure_time: Exposure time for images.

        Returns:
            Tuple of obtained best focus value and its uncertainty.

        Raises:
            ValueError: If focus could not be obtained.
        """
        ...

    @abstractmethod
    async def auto_focus_status(self, **kwargs: Any) -> Dict[str, Any]:
        """Returns current status of auto focus.

        Returned dictionary contains a list of focus/fwhm pairs in X and Y direction.

        Returns:
            Dictionary with current status.
        """
        ...


__all__ = ["IAutoFocus"]



from __future__ import annotations
import copy
import io
from typing import TypeVar, Optional, Type, Dict, Any, cast

import numpy as np
from astropy.io import fits
from astropy.io.fits import table_to_hdu, ImageHDU
from astropy.table import Table
from astropy.nddata import CCDData, StdDevUncertainty
from numpy.typing import NDArray

from pyobs.utils.fits import FilenameFormatter

MetaClass = TypeVar("MetaClass")


class Image:
    """Image class."""

    __module__ = "pyobs.images"

    def __init__(
        self,
        data: Optional[NDArray[Any]] = None,
        header: Optional[fits.Header] = None,
        mask: Optional[NDArray[Any]] = None,
        uncertainty: Optional[NDArray[Any]] = None,
        catalog: Optional[Table] = None,
        meta: Optional[Dict[Any, Any]] = None,
        *args: Any,
        **kwargs: Any,
    ):
        """Init a new image.

        Args:
            data: Numpy array containing data for image.
            header: Header for the new image.
            mask: Mask for the image.
            uncertainty: Uncertainty image.
            catalog: Catalog table.
            meta: Dictionary with meta information (note: not preserved in I/O operations!).
        """

        # store
        self.data = data
        self.header = fits.Header() if header is None else header.copy()
        self.mask = None if mask is None else mask.copy()
        self.uncertainty = None if uncertainty is None else uncertainty.copy()
        self.catalog = None if catalog is None else catalog.copy()
        self.meta = {} if meta is None else copy.deepcopy(meta)

        # add basic header stuff
        if data is not None:
            self.header["NAXIS1"] = data.shape[1]
            self.header["NAXIS2"] = data.shape[0]

    @classmethod
    def from_bytes(cls, data: bytes) -> Image:
        """Create Image from a bytes array containing a FITS file.

        Args:
            data: Bytes array to create image from.

        Returns:
            The new image.
        """

        # create hdu
        with io.BytesIO(data) as bio:
            # read whole file
            data = fits.open(bio, memmap=False, lazy_load_hdus=False)

            # load image
            return cls._from_hdu_list(data)

    @classmethod
    def from_file(cls, filename: str) -> Image:
        """Create image from FITS file.

        Args:
            filename: Name of file to load image from.

        Returns:
            New image.
        """

        # open file
        data = fits.open(filename, memmap=False, lazy_load_hdus=False)

        # load image
        image = cls._from_hdu_list(data)

        # close file
        data.close()
        return image

    @classmethod
    def from_ccddata(cls, data: CCDData) -> Image:
        """Create image from astropy.CCDData.

        Args:
            data: CCDData to create image from.

        Returns:
            New image.
        """

        # create image and assign data
        image = Image(
            data=data.data.astype(np.float32),
            header=data.header,
            mask=None if data.mask is None else data.mask,
            uncertainty=None if data.uncertainty is None else data.uncertainty.array.astype(np.float32),
        )
        return image

    @classmethod
    def _from_hdu_list(cls, data: fits.HDUList) -> "Image":
        """Load Image from HDU list.

        Args:
            data: HDU list.

        Returns:
            Image.
        """

        # create image
        image = cls()

        # find HDU with image data
        for hdu in data:
            if (
                isinstance(hdu, fits.PrimaryHDU)
                and hdu.header["NAXIS"] > 0
                or isinstance(hdu, fits.ImageHDU)
                and hdu.name == "SCI"
                or isinstance(hdu, fits.CompImageHDU)
            ):
                # found image HDU
                image_hdu = hdu
                break
        else:
            raise ValueError("Could not find HDU with main image.")

        # get data
        image.data = image_hdu.data
        image.header = image_hdu.header

        # mask
        if "MASK" in data:
            image.mask = data["MASK"].data

        # uncertainties
        if "UNCERT" in data:
            image.uncertainty = data["UNCERT"].data

        # catalog
        if "CAT" in data:
            image.catalog = Table(data["CAT"].data)

        # finished
        return image

    @property
    def unit(self) -> str:
        """Returns units of pixels in image."""
        return str(self.header["BUNIT"]).lower() if "BUNIT" in self.header else "adu"

    def __deepcopy__(self) -> Image:
        """Returns a shallow copy of this image."""
        return self.copy()

    def copy(self) -> Image:
        """Returns a copy of this image."""
        return Image(
            data=self.data,
            header=self.header,
            mask=self.mask,
            uncertainty=self.uncertainty,
            catalog=self.catalog,
            meta=self.meta,
        )

    def __truediv__(self, other: "Image") -> "Image":
        """Divides this image by other."""
        img = self.copy()
        if img.data is None or other.data is None:
            raise ValueError("One image in division is None.")
        img.data /= other.data
        return img

    def writeto(self, f: Any, *args: Any, **kwargs: Any) -> None:
        """Write image as FITS to given file object.

        Args:
            f: File object to write to.
        """

        # create HDU list
        hdu_list = fits.HDUList([])

        # create image HDU
        hdu = fits.PrimaryHDU(self.data, header=self.header)
        hdu_list.append(hdu)

        # catalog?
        if self.catalog is not None:
            hdu = table_to_hdu(self.catalog)
            hdu.name = "CAT"
            hdu_list.append(hdu)

        # mask?
        if self.mask is not None:
            hdu = ImageHDU(self.mask.astype(np.uint8))
            hdu.name = "MASK"
            hdu_list.append(hdu)

        # errors?
        if self.uncertainty is not None:
            hdu = ImageHDU(self.uncertainty.data)
            hdu.name = "UNCERT"
            hdu_list.append(hdu)

        # write it
        hdu_list.writeto(f, *args, **kwargs)

    def to_bytes(self) -> bytes:
        """Write to a bytes array and return it."""
        with io.BytesIO() as bio:
            self.writeto(bio)
            return bio.getvalue()

    def write_catalog(self, f: Any, *args: Any, **kwargs: Any) -> None:
        """Write catalog to file object."""
        if self.catalog is None:
            return

        # create HDU and write it
        table_to_hdu(self.catalog).writeto(f, *args, **kwargs)

    def to_ccddata(self) -> CCDData:
        """Convert Image to CCDData"""
        return CCDData(
            data=self.data,
            meta=self.header,
            mask=None if self.mask is None else self.mask,
            uncertainty=None if self.uncertainty is None else StdDevUncertainty(self.uncertainty),
            unit="adu",
        )

    def format_filename(self, formatter: FilenameFormatter) -> str:
        """Format filename with given formatter."""
        self.header["FNAME"] = formatter(self.header)
        return str(self.header["FNAME"])

    @property
    def pixel_scale(self) -> Optional[float]:
        """Returns pixel scale in arcsec/pixel."""
        if "CD1_1" in self.header:
            return abs(float(self.header["CD1_1"])) * 3600.0
        elif "CDELT1" in self.header:
            return abs(float(self.header["CDELT1"])) * 3600.0
        else:
            return None

    def to_jpeg(self, vmin: Optional[float] = None, vmax: Optional[float] = None) -> bytes:
        """Returns a JPEG image created from this image.

        Returns:
            The image.
        """

        # import PIL Image
        import PIL.Image

        # copy data
        data: NDArray[Any] = np.copy(self.data)  # type: ignore
        if data is None:
            raise ValueError("No data in image.")

        # no vmin/vmax?
        if vmin is None or vmax is None:
            flattened = sorted(data.flatten())
            vmin = flattened[int(0.05 * len(flattened))]
            vmax = flattened[int(0.95 * len(flattened))]
            if vmin is None or vmax is None:
                raise ValueError("Could not determine vmin/vmax.")

        # Clip data to brightness limits
        data[data > vmax] = vmax
        data[data < vmin] = vmin

        # Scale data to range [0, 1]
        data = (data - vmin) / (vmax - vmin)

        # Convert to 8-bit integer
        data = (255 * data).astype(np.uint8)

        # Invert y axis
        data = data[::-1, :]

        # create image from data array
        image = PIL.Image.fromarray(data, "L")
        with io.BytesIO() as bio:
            image.save(bio, format="jpeg")
            return bio.getvalue()

    def set_meta(self, meta: Any) -> None:
        """Sets meta information, storing it under it class.

        Note that it is possible to store, e.g., strings, but they would be stored as img.meta[str] and be overwritten
        with every new string, which is probably not what you want. Use the img.meta dict directly for this and
        set_meta/get_meta only for class-based data.

        Args:
            meta: Meta information to store.
        """

        # store it
        self.meta[meta.__class__] = meta

    def has_meta(self, meta_class: Type[MetaClass]) -> bool:
        """Whether meta exists."""
        return meta_class in self.meta

    def get_meta(self, meta_class: Type[MetaClass]) -> MetaClass:
        """Returns meta information, assuming that it is stored under the class of the object.

        Args:
            meta_class: Class to return meta information for.

        Returns:
            Meta information of the given class.
        """
        # return default?
        if meta_class not in self.meta:
            raise ValueError("Meta value not found.")

        # correct type?
        if not isinstance(self.meta[meta_class], meta_class):
            raise ValueError("Stored meta information is of wrong type.")

        # return it
        return cast(MetaClass, self.meta[meta_class])

    def get_meta_safe(self, meta_class: Type[MetaClass], default: Optional[MetaClass] = None) -> Optional[MetaClass]:
        """Calls get_meta in a safe way and returns default value in case of an exception."""

        try:
            return self.get_meta(meta_class)
        except:
            return default


__all__ = ["Image"]



import asyncio
import io
import logging
from enum import Enum
from functools import partial
from inspect import Parameter
from pprint import pprint
from typing import Any, Optional, List

from telegram import InlineKeyboardButton, InlineKeyboardMarkup, Update
from telegram.ext import CallbackQueryHandler, CallbackContext

from pyobs.modules import Module
from pyobs.events import LogEvent, Event

log = logging.getLogger(__name__)


class TelegramUserState(Enum):
    IDLE = (0,)
    AUTH = (1,)
    EXEC_MODULE = (2,)
    EXEC_METHOD = (3,)
    EXEC_PARAMS = (4,)
    LOG_LEVEL = 5


class Telegram(Module):
    """A telegram bot."""

    __module__ = "pyobs.modules.utils"

    def __init__(self, token: str, password: str, allow_new_users: bool = True, **kwargs: Any):
        """Initialize a new bot.

        Args:
            token: The telegram API token.
            password: Password for users to log in.
            allow_new_users: Whether new users are allowed to connect.

        """
        Module.__init__(self, **kwargs)
        from telegram.ext import Updater

        # store
        self._token = token
        self._password = password
        self._allow_new_users = allow_new_users
        self._updater: Optional[Updater] = None
        self._message_queue = asyncio.Queue()
        self._loop = None

        # get log levels
        self._log_levels = {
            logging.getLevelName(x): x for x in range(1, 101) if not logging.getLevelName(x).startswith("Level")
        }

        # thread
        self.add_background_task(self._log_sender_thread)

    async def open(self) -> None:
        """Open module."""
        from telegram.ext import CommandHandler, MessageHandler, Filters, Updater

        await Module.open(self)
        self._loop = asyncio.get_running_loop()

        # get dispatcher
        self._updater = Updater(token=self._token)
        dispatcher = self._updater.dispatcher  # type: ignore

        # add command handler
        dispatcher.add_handler(CommandHandler("start", self._command_start))
        dispatcher.add_handler(CommandHandler("exec", self._command_exec))
        dispatcher.add_handler(CommandHandler("modules", self._command_modules))
        dispatcher.add_handler(CommandHandler("loglevel", self._command_loglevel))

        # add text handler
        echo_handler = MessageHandler(Filters.text & (~Filters.command), self._process_message)
        dispatcher.add_handler(echo_handler)

        # add callback handler for buttons
        dispatcher.add_handler(CallbackQueryHandler(self._handle_buttons))

        # load storage file
        try:
            dispatcher.bot_data["storage"] = await self.vfs.read_yaml("/pyobs/telegram.yaml")
        except FileNotFoundError:
            dispatcher.bot_data["storage"] = {}

        # start polling
        self._updater.start_polling(poll_interval=0.1)

        # listen to log events
        await self.comm.register_event(LogEvent, self._process_log_entry)

    async def close(self) -> None:
        """Close module."""
        await Module.close(self)

        # stop telegram
        if self._updater is not None:
            self._updater.stop()

    async def _save_storage(self, context: CallbackContext) -> None:
        """Save storage file.

        Args:
            context: Telegram context.
        """
        await self.vfs.write_yaml("/pyobs/telegram.yaml", context.bot_data["storage"])

    @staticmethod
    def _is_user_authorized(context: CallbackContext, user_id: int) -> bool:
        """Is user authorized?

        Args:
            context: Telegram context.
            user_id: ID of user.

        Returns:
            Whether user is known and authorized to give commands.
        """
        s = context.bot_data["storage"]
        return "users" in s and user_id in s["users"]

    def _store_user(self, context: CallbackContext, user_id: int, name: str) -> None:
        """Store new user in auth database.

        Args:
            context: Telegram context.
            user_id: ID of user.
            name: Name of user.
        """

        # add user
        s = context.bot_data["storage"]
        if "users" not in s:
            s["users"] = {}
        s["users"][user_id] = {"name": name, "loglevel": None}
        asyncio.run_coroutine_threadsafe(self._save_storage(context), self._loop)

    def _command_start(self, update: Update, context: CallbackContext) -> None:
        """Handle /start command.

        Args:
            update: Message to process.
            context: Telegram context.
        """

        if context.user_data is None:
            raise ValueError("No user data in context.")

        # is user already known?
        if self._is_user_authorized(context, update.message.from_user.id):
            # welcome him back
            context.bot.send_message(
                chat_id=update.effective_chat.id, text="Welcome back %s!" % update.message.from_user.first_name
            )

        else:
            # do we allow for new users?
            if self._allow_new_users:
                # go to AUTH state
                context.bot.send_message(chat_id=update.effective_chat.id, text="Password?")
                context.user_data["state"] = TelegramUserState.AUTH

            else:
                # show message
                context.bot.send_message(chat_id=update.effective_chat.id, text="No new users allowed in the system.")

    def _command_exec(self, update: Update, context: CallbackContext) -> None:
        """Handle /exec command.

        Args:
            update: Message to process.
            context: Telegram context.
        """

        if context.user_data is None:
            raise ValueError("No user data in context.")

        # not logged in?
        if not self._is_user_authorized(context, update.message.from_user.id):
            context.bot.send_message(chat_id=update.effective_chat.id, text="Not logged in, use /start.")
            return

        # create buttons for all modules
        keyboard = [[InlineKeyboardButton(c, callback_data=c)] for c in self.comm.clients] + [
            [InlineKeyboardButton("Cancel", callback_data="cancel")]
        ]
        reply_markup = InlineKeyboardMarkup(keyboard)
        update.message.reply_text("Please choose module:", reply_markup=reply_markup)

        # go to EXEC_MODULE state
        context.user_data["state"] = TelegramUserState.EXEC_MODULE

    @staticmethod
    def _reset_state(context: CallbackContext) -> None:
        """Reset state."""
        if context.user_data is None:
            raise ValueError("No user data in context.")
        context.user_data["state"] = TelegramUserState.IDLE
        context.user_data["method"] = None
        context.user_data["params"] = []
        context.user_data["exec_query"] = None

    def _handle_buttons(self, update: Update, context: CallbackContext) -> None:
        """Handle click on buttons.

        Args:
            update: Message to process.
            context: Telegram context.
        """
        asyncio.run_coroutine_threadsafe(self._handle_buttons_async(update, context), self._loop)

    async def _handle_buttons_async(self, update: Update, context: CallbackContext) -> None:
        """Handle click on buttons.

        Args:
            update: Message to process.
            context: Telegram context.
        """
        if context.user_data is None:
            raise ValueError("No user data in context.")

        # get query
        query = update.callback_query

        # not logged in?
        if not self._is_user_authorized(context, query.from_user.id):
            query.edit_message_text(text="Not logged in, use /start.")
            return

        # CallbackQueries need to be answered, even if no notification to the user is needed
        # Some clients may have trouble otherwise. See https://core.telegram.org/bots/api#callbackquery
        query.answer()

        # cancel?
        if query.data == "cancel":
            # reset state
            self._reset_state(context)

            # remove markup
            query.edit_message_text("Canceled.")

        # what state are we in?
        if context.user_data["state"] == TelegramUserState.EXEC_MODULE:
            # get proxy for selected module
            proxy = await self.proxy(query.data)

            # show buttons for all modules
            keyboard = [
                [InlineKeyboardButton(m, callback_data="%s.%s" % (query.data, m))] for m in proxy.method_names
            ] + [[InlineKeyboardButton("Cancel", callback_data="cancel")]]
            reply_markup = InlineKeyboardMarkup(keyboard)
            query.edit_message_text(text="Chose method in %s:" % query.data)
            query.edit_message_reply_markup(reply_markup)

            # change to EXEC_MEHOD state
            context.user_data["state"] = TelegramUserState.EXEC_METHOD

        elif context.user_data["state"] == TelegramUserState.EXEC_METHOD:
            # init command
            context.user_data["method"] = query.data
            context.user_data["params"] = []
            context.user_data["exec_query_message"] = update.callback_query.message.message_id
            context.user_data["exec_query_chat"] = update.callback_query.message.chat_id

            # show buttons for all methods
            keyboard = [[InlineKeyboardButton("Cancel", callback_data="cancel")]]
            reply_markup = InlineKeyboardMarkup(keyboard)
            query.edit_message_text(text="Executing %s..." % query.data)
            query.edit_message_reply_markup(reply_markup)

            # go to EXEC_PARAMS state
            context.user_data["state"] = TelegramUserState.EXEC_PARAMS

            # see whether this command runs without parameters
            self._handle_params(update, context)

        elif context.user_data["state"] == TelegramUserState.LOG_LEVEL:
            # set log level
            s = context.bot_data["storage"]
            s["users"][query.from_user.id]["loglevel"] = query.data
            await self._save_storage(context)

            # change to IDLE state
            query.edit_message_text(text="Changed log level to %s." % query.data)
            context.user_data["state"] = TelegramUserState.IDLE

    def _handle_params(self, update: Update, context: CallbackContext) -> None:
        """Handle input of params when in EXEC_PARAMS state.

        Args:
            update: Message to process.
            context: Telegram context.
        """
        asyncio.run_coroutine_threadsafe(self._handle_params_async(update, context), self._loop)

    async def _handle_params_async(self, update: Update, context: CallbackContext) -> None:
        """Handle input of params when in EXEC_PARAMS state.

        Args:
            update: Message to process.
            context: Telegram context.
        """

        if context.user_data is None:
            raise ValueError("No user data in context.")

        # wrong state?
        if context.user_data["state"] != TelegramUserState.EXEC_PARAMS:
            return

        # get proxy and method signature
        module, method = context.user_data["method"].split(".")
        proxy = await self.proxy(module)
        signature = proxy.signature(method)

        # get list of parameters
        params = [
            (name, param) for name, param in signature.parameters.items() if name not in ["self", "args", "kwargs"]
        ]

        # append param, if new one was received
        if update.message is not None:
            # get type of new parameter
            param_type = params[len(context.user_data["params"])][1].annotation

            # cast it
            value = param_type(update.message.text)

            # append it
            context.user_data["params"].append(value)

        # got enough params?
        nparams = len(context.user_data["params"])
        if nparams >= len(params):
            # ID for command?
            if "ncalls" not in context.user_data:
                context.user_data["ncalls"] = 0
            context.user_data["ncalls"] += 1
            call_id = context.user_data["ncalls"]

            # remove cancel button, and show command
            msg = "Executing %s..." % context.user_data["method"]
            context.bot.edit_message_text(
                text=msg,
                message_id=context.user_data["exec_query_message"],
                chat_id=context.user_data["exec_query_chat"],
            )

            # send message
            command = (
                context.user_data["method"]
                + "("
                + ", ".join(['"%s"' % p if isinstance(p, str) else str(p) for p in context.user_data["params"]])
                + ")"
            )
            context.bot.send_message(chat_id=update.effective_chat.id, text="Executing #%d:\n%s" % (call_id, command))

            # start call
            asyncio.create_task(
                self._call_method(
                    context, update.effective_chat.id, call_id, context.user_data["method"], context.user_data["params"]
                )
            )

            # reset
            self._reset_state(context)

        else:
            # no, print next one
            next_param: Parameter = params[nparams][1]

            # format it
            message = "Value for " + next_param.name
            if next_param.annotation is not None:
                message += ": " + next_param.annotation.__name__
            if next_param.default != Parameter.empty:
                message += " = " + str(next_param.default)
            message += "?"

            # send it
            context.bot.send_message(chat_id=update.effective_chat.id, text=message)

    async def _call_method(
        self, context: CallbackContext, chat_id: int, call_id: int, method: str, params: List[Any]
    ) -> None:
        """

        Args:
            context: Telegram context.
            chat_id: Telegram chat ID.
            call_id: ID of command.
            method: Method to call as <module>.<method>
            params: List of parameters.
        """
        # get proxy
        module, method_name = method.split(".")
        proxy = await self.proxy(module)

        # call it
        func = getattr(proxy, method_name)
        response = await func(*params)

        # set message
        if response is None:
            message = "Finished #%d." % call_id

        else:
            # format message
            with io.StringIO() as sio:
                # format response
                pprint(response, stream=sio, indent=2)
                message = "Finished #%d:\n%s" % (call_id, sio.getvalue())

        # send reply
        context.bot.send_message(chat_id=chat_id, text=message)

    def _process_message(self, update: Update, context: CallbackContext) -> None:
        """Handle normal text messages, e.g. for login or method parameters.

        Args:
            update: Telegram message.
            context: Telegram context.
        """

        if context.user_data is None:
            raise ValueError("No user data in context.")

        # what state is user in?
        if context.user_data["state"] == TelegramUserState.AUTH:
            # AUTH, so we expect a password. Is it valid?
            if update.message.text == self._password:
                # Yes, successful AUTH
                context.bot.send_message(chat_id=update.effective_chat.id, text="AUTH successful.")
                context.user_data["state"] = TelegramUserState.IDLE
                self._store_user(context, update.message.from_user.id, update.message.from_user.first_name)

            else:
                context.bot.send_message(chat_id=update.effective_chat.id, text="AUTH failed, try again.")

        elif context.user_data["state"] == TelegramUserState.EXEC_PARAMS:
            # we're expecting params, so handle them
            self._handle_params(update, context)

    def _command_modules(self, update: Update, context: CallbackContext) -> None:
        """Handle /modules command that shows list of modules.

        Args:
            update: Message to process.
            context: Telegram context.
        """

        # not logged in?
        if not self._is_user_authorized(context, update.message.from_user.id):
            context.bot.send_message(chat_id=update.effective_chat.id, text="Not logged in, use /start.")
            return

        # list all modules
        message = "Available modules:\n" + "\n".join(["- " + c for c in self.comm.clients])
        context.bot.send_message(chat_id=update.effective_chat.id, text=message)

    def _command_loglevel(self, update: Update, context: CallbackContext) -> None:
        """Handle /loglevel command that sets the log level

        Args:
            update: Message to process.
            context: Telegram context.
        """

        if context.user_data is None:
            raise ValueError("No user data in context.")

        # not logged in?
        if not self._is_user_authorized(context, update.message.from_user.id):
            context.bot.send_message(chat_id=update.effective_chat.id, text="Not logged in, use /start.")
            return

        # set state
        context.user_data["state"] = TelegramUserState.LOG_LEVEL

        # get current level
        s = context.bot_data["storage"]
        current_level = s["users"][update.message.from_user.id]["loglevel"]

        # create buttons for all log levels
        keyboard = [[InlineKeyboardButton(level, callback_data=level)] for level in self._log_levels.keys()] + [
            [InlineKeyboardButton("Cancel", callback_data="cancel")]
        ]
        reply_markup = InlineKeyboardMarkup(keyboard)
        update.message.reply_text(
            "Current log level: %s\nPlease choose new log level:" % current_level, reply_markup=reply_markup
        )

    async def _process_log_entry(self, entry: Event, sender: str) -> bool:
        """Process a new log entry.

        Args:
            entry: The log event.
            sender: Name of sender.
        """
        if not isinstance(entry, LogEvent):
            return False

        # get numerical value for log level
        level = self._log_levels[entry.level]

        # build log message
        message = "(%s) %s: %s" % (entry.level, sender, entry.message)

        # get storage
        if self._updater is None:
            raise ValueError("No update initialised.")
        s = self._updater.dispatcher.bot_data["storage"]  # type: ignore

        # loop users
        for user_id, user in s["users"].items():
            # get user log level
            user_level = self._log_levels[user["loglevel"]] if user["loglevel"] in self._log_levels else 100

            # is it larger than the log entry level?
            if level >= user_level:
                # queue message
                self._message_queue.put_nowait((user_id, message))

        return True

    async def _log_sender_thread(self) -> None:
        """Thread for sending messages."""

        loop = asyncio.get_running_loop()
        while True:
            # get next entry
            user_id, message = await self._message_queue.get()

            # send message
            try:
                if self._updater is None:
                    raise ValueError("No update initialised.")
                await loop.run_in_executor(None, partial(self._updater.bot.send_message, chat_id=user_id, text=message))

            except Exception:
                # something went wrong, sleep a little and queue message again
                await asyncio.sleep(10)
                await self._message_queue.put((user_id, message))


__all__ = ["Telegram"]



import os
import re


def pre_process_yaml(config: str) -> str:
    # read config
    with open(config, "r") as f:
        content = f.read()

    # get path of config
    path = os.path.dirname(os.path.abspath(config))

    # find all include statements
    matches = re.findall("(\{include (.*)\})", content)
    for match, filename in matches:
        content = content.replace(match, pre_process_yaml(os.path.join(path, filename)))

    # return new yaml
    return content


__all__ = ["pre_process_yaml"]



import logging
import random
from typing import Tuple, Any, Optional, List, Dict

import numpy as np
import pandas as pd
from astropy.coordinates import SkyCoord
import astropy.units as u

from pyobs.interfaces import IAcquisition, ITelescope
from pyobs.modules import Module
from pyobs.utils import exceptions as exc
from pyobs.interfaces import IAutonomous
from pyobs.utils.time import Time

log = logging.getLogger(__name__)


class PointingSeries(Module, IAutonomous):
    """Module for running pointing series."""

    __module__ = "pyobs.modules.robotic"

    def __init__(
        self,
        alt_range: Tuple[float, float] = (30.0, 85.0),
        num_alt: int = 8,
        az_range: Tuple[float, float] = (0.0, 360.0),
        num_az: int = 24,
        dec_range: Tuple[float, float] = (-80.0, 80.0),
        min_moon_dist: float = 15.0,
        finish: int = 90,
        exp_time: float = 1.0,
        acquisition: str = "acquisition",
        telescope: str = "telescope",
        **kwargs: Any,
    ):
        """Initialize a new auto focus system.

        Args:
            alt_range: Range in degrees to use in altitude.
            num_alt: Number of altitude points to create on grid.
            az_range: Range in degrees to use in azimuth.
            num_az: Number of azimuth points to create on grid.
            dec_range: Range in declination in degrees to use.
            min_moon_dist: Minimum moon distance in degrees.
            finish: When this number in percent of points have been finished, terminate mastermind.
            exp_time: Exposure time in secs.
            acquisition: IAcquisition unit to use.
            telescope: ITelescope unit to use.
        """
        Module.__init__(self, **kwargs)

        # store
        self._alt_range = tuple(alt_range)
        self._num_alt = num_alt
        self._az_range = tuple(az_range)
        self._num_az = num_az
        self._dec_range = dec_range
        self._min_moon_dist = min_moon_dist
        self._finish = 1.0 - finish / 100.0
        self._exp_time = exp_time
        self._acquisition = acquisition
        self._telescope = telescope

        # if Az range is [0, 360], we got north double, so remove one step
        if self._az_range == (0.0, 360.0):
            self._az_range = (0.0, 360.0 - 360.0 / self._num_az)

        # add thread func
        self.add_background_task(self._run_thread, False)

    async def start(self, **kwargs: Any) -> None:
        """Starts a service."""
        pass

    async def stop(self, **kwargs: Any) -> None:
        """Stops a service."""
        pass

    async def is_running(self, **kwargs: Any) -> bool:
        """Whether a service is running."""
        return True

    async def _run_thread(self) -> None:
        """Run a pointing series."""

        # create grid
        grid: Dict[str, List[Any]] = {"alt": [], "az": [], "done": []}
        for az in np.linspace(self._az_range[0], self._az_range[1], self._num_az):
            for alt in np.linspace(self._alt_range[0], self._alt_range[1], self._num_alt):
                grid["alt"] += [alt]
                grid["az"] += [az]
                grid["done"] += [False]

        # to dataframe
        pd_grid = pd.DataFrame(grid).set_index(["alt", "az"])

        # get acquisition and telescope units
        acquisition = await self.proxy(self._acquisition, IAcquisition)
        telescope = await self.proxy(self._telescope, ITelescope)

        # check observer
        if self.observer is None:
            raise ValueError("No observer given.")

        # loop until finished
        while True:
            # get all entries without offset measurements
            todo = list(pd_grid[~pd_grid["done"]].index)
            if len(todo) / len(pd_grid) < self._finish:
                log.info("Finished.")
                break
            log.info("Grid points left to do: %d", len(todo))

            # get moon
            moon = self.observer.moon_altaz(Time.now())

            # try to find a good point
            while True:
                # pick a random index and remove from list
                alt, az = random.sample(todo, 1)[0]
                todo.remove((alt, az))
                altaz = SkyCoord(
                    alt=alt * u.deg, az=az * u.deg, frame="altaz", obstime=Time.now(), location=self.observer.location
                )

                # get RA/Dec
                radec = altaz.icrs

                # moon far enough away?
                if altaz.separation(moon).degree >= self._min_moon_dist:
                    # yep, are we in declination range?
                    if self._dec_range[0] <= radec.dec.degree < self._dec_range[1]:
                        # yep, break here, we found our target
                        break

                # to do list empty?
                if len(todo) == 0:
                    # could not find a grid point
                    log.info("Could not find a suitable grid point, resetting todo list for next entry...")
                    todo = list(pd_grid.index)
                    continue

            # log finding
            log.info("Picked grid point at Alt=%.2f, Az=%.2f (%s).", alt, az, radec.to_string("hmsdms"))

            # acquire target and process result
            try:
                # move telescope
                await telescope.move_radec(float(radec.ra.degree), float(radec.dec.degree))

                # acquire target
                acq = await acquisition.acquire_target()

                #  process result
                if acq is not None:
                    await self._process_acquisition(**acq)

            except (ValueError, exc.RemoteError):
                log.info("Could not acquire target.")
                continue

            # finished
            pd_grid.loc[alt, az] = True

        # finished
        log.info("Pointing series finished.")

    async def _process_acquisition(
        self,
        datetime: str,
        ra: float,
        dec: float,
        alt: float,
        az: float,
        off_ra: Optional[float] = None,
        off_dec: Optional[float] = None,
        off_alt: Optional[float] = None,
        off_az: Optional[float] = None,
    ) -> None:
        """Process the result of the acquisition. Either ra_off/dec_off or alt_off/az_off must be given.

        Args:
            datetime: Date and time of observation.
            ra: Right ascension without offsets at destination.
            dec: Declination without offsets at destination.
            alt: Altitude without offsets at destination.
            az: Azimuth without offsets at destination.
            off_ra: Found RA offset.
            off_dec: Found Dec offset.
            off_alt: Found Alt offset.
            off_az: Found Az offset.
        """
        pass


__all__ = ["PointingSeries"]



"""
TODO: write doc
"""
__title__ = "Time"

import datetime
from typing import cast

import astropy.time
import astropy.units as u
import pytz
from astroplan import Observer


class Time(astropy.time.Time):  # type: ignore
    """Hashable Time class."""

    _now_offset = astropy.time.TimeDelta(0 * u.second)

    def __hash__(self) -> int:
        if self.ndim != 0:
            raise TypeError("unhashable type: '{}'".format(self.__class__.__name__))
        return hash((self.jd1, self.jd2, self.scale))

    @classmethod
    def set_offset_to_now(cls, delta: astropy.time.TimeDelta) -> None:
        cls._now_offset = delta

    @classmethod
    def now(cls) -> "Time":
        """
        Creates a new object corresponding to the instant in time this
        method is called.

        .. note::
            "Now" is determined using the `~datetime.datetime.utcnow`
            function, so its accuracy and precision is determined by that
            function.  Generally that means it is set by the accuracy of
            your system clock.

        Returns:
            A new `Time` object (or a subclass of `Time` if this is called from
            such a subclass) at the current time.
        """
        # call `utcnow` immediately to be sure it's ASAP
        dtnow = datetime.datetime.utcnow()
        return cast(Time, Time(val=dtnow, format="datetime", scale="utc") + Time._now_offset)

    def night_obs(self, observer: Observer) -> datetime.date:
        """Returns the night for this time, i.e. the date of the start of the current night.

        Args:
            observer: Observer object to use.

        Returns:
            Night for this time.
        """

        # convert to datetime
        time = self.datetime

        # get local datetime
        utc_dt = pytz.utc.localize(time)
        loc_dt = utc_dt.astimezone(observer.timezone)

        # get night
        if loc_dt.hour < 15:
            loc_dt += datetime.timedelta(days=-1)
        return loc_dt.date()


__all__ = ["Time"]



import logging
from abc import ABCMeta
from typing import List, Dict, Tuple, Any, Optional

from pyobs.interfaces import IRoof, IFitsHeaderBefore
from pyobs.modules import Module
from pyobs.mixins import MotionStatusMixin, WeatherAwareMixin
from pyobs.utils.enums import MotionStatus

log = logging.getLogger(__name__)


class BaseRoof(WeatherAwareMixin, MotionStatusMixin, IRoof, IFitsHeaderBefore, Module, metaclass=ABCMeta):
    """Base class for roofs."""

    __module__ = "pyobs.modules.roof"

    def __init__(self, **kwargs: Any):
        """Initialize a new base roof."""
        Module.__init__(self, **kwargs)

        # init mixins
        WeatherAwareMixin.__init__(self, **kwargs)
        MotionStatusMixin.__init__(self, **kwargs)

    async def open(self) -> None:
        """Open module."""
        await Module.open(self)

        # open mixins
        await WeatherAwareMixin.open(self)
        await MotionStatusMixin.open(self)

    async def get_fits_header_before(
        self, namespaces: Optional[List[str]] = None, **kwargs: Any
    ) -> Dict[str, Tuple[Any, str]]:
        """Returns FITS header for the current status of this module.

        Args:
            namespaces: If given, only return FITS headers for the given namespaces.

        Returns:
            Dictionary containing FITS headers.
        """
        return {
            "ROOF-OPN": (
                await self.get_motion_status() in [MotionStatus.POSITIONED, MotionStatus.TRACKING],
                "True for open, false for closed roof",
            )
        }

    async def is_ready(self, **kwargs: Any) -> bool:
        """Returns the device is "ready", whatever that means for the specific device.

        Returns:
            True, if roof is open.
        """

        # check that motion is not in one of the states listed below
        return await self.get_motion_status() not in [
            MotionStatus.PARKED,
            MotionStatus.INITIALIZING,
            MotionStatus.PARKING,
            MotionStatus.ERROR,
            MotionStatus.UNKNOWN,
        ]


__all__ = ["BaseRoof"]



from abc import ABCMeta, abstractmethod
from typing import Any, Dict

from .IRunning import IRunning


class IAcquisition(IRunning, metaclass=ABCMeta):
    """The module can acquire a target, usually by accessing a telescope and a camera."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def acquire_target(self, **kwargs: Any) -> Dict[str, Any]:
        """Acquire target at given coordinates.

        If no RA/Dec are given, start from current position. Might not work for some implementations that require
        coordinates.

        Returns:
            A dictionary with entries for datetime, ra, dec, alt, az, and either off_ra, off_dec or off_alt, off_az.

        Raises:
            ValueError: If target could not be acquired.
        """
        ...


__all__ = ["IAcquisition"]



from inspect import BoundArguments, Signature, Parameter
from enum import Enum
from typing import Any, get_origin
import xml.sax.saxutils


def cast_bound_arguments_to_simple(bound_arguments: BoundArguments):
    """Cast the requested parameters, which are of simple types, to the types required by the method.

    Args:
        bound_arguments: Incoming parameters.
    """
    # loop all arguments
    for key, value in bound_arguments.arguments.items():
        # special cases
        if isinstance(value, str):
            # escape strings
            bound_arguments.arguments[key] = xml.sax.saxutils.escape(value)
        elif isinstance(value, Enum):
            # get value of enum
            bound_arguments.arguments[key] = value.value


def cast_bound_arguments_to_real(bound_arguments: BoundArguments, signature: Signature):
    """Cast the requested parameters to simple types.

    Args:
        bound_arguments: Incoming parameters.
        signature: Signature of method.
    """
    # loop all arguments
    for key, value in bound_arguments.arguments.items():
        # get type of parameter
        annotation = signature.parameters[key].annotation

        # special cases
        if value is None or annotation == Parameter.empty or annotation == Any:
            # if value is None or no annotation is given, just copy it
            bound_arguments.arguments[key] = value
        elif annotation == Enum:
            # cast to enum
            bound_arguments.arguments[key] = value if annotation == Parameter.empty else annotation(value)
        elif annotation == str:
            # unescape strings
            bound_arguments.arguments[key] = xml.sax.saxutils.unescape(value)
        else:
            # cast to type
            bound_arguments.arguments[key] = annotation(value)


def cast_response_to_real(response: Any, signature: Signature) -> Any:
    """Cast a response from simple to the method's real types.

    Args:
        response: Response of method call.
        signature: Signature of method.

    Returns:
        Same as input response, but with only simple types.
    """

    # get return annotation
    annotation = signature.return_annotation

    # TODO: For future Python versions (3.9?)
    # - handle dicts and tuples

    # any annotations?
    if response is None or annotation is None or annotation == Parameter.empty or annotation == Any:
        # no response or no annotation at all or Any
        return response
    elif (get_origin(annotation) == tuple) or isinstance(annotation, tuple):
        # parse tuple
        # return tuple([None if res is None else annot(res) for res, annot in zip(response, annotation)])
        return response
    elif (get_origin(annotation) == list) or isinstance(annotation, list):
        # parse list
        return response
    elif (get_origin(annotation) == dict) or isinstance(annotation, dict):
        # just return it
        return response
    else:
        # type cast response
        return annotation(response)


def cast_response_to_simple(response: Any) -> Any:
    """Cast a response from a method to only simple types.

    Args:
        response: Response of method call.

    Returns:
        Same as input response, but with only simple types.
    """

    # tuple, enum or something else
    if isinstance(response, tuple):
        return tuple([cast_response_to_simple(r) for r in response])
    elif isinstance(response, Enum):
        return response.value
    else:
        return response



from abc import ABCMeta, abstractmethod
from typing import List, Dict, Tuple, Any, Optional

from .interface import Interface


class IFitsHeaderAfter(Interface, metaclass=ABCMeta):
    """The module provides some additional header entries for FITS headers after some event (usually the end of the
    exposure)."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def get_fits_header_after(
        self, namespaces: Optional[List[str]] = None, **kwargs: Any
    ) -> Dict[str, Tuple[Any, str]]:
        """Returns FITS header for the current status of this module.

        Args:
            namespaces: If given, only return FITS headers for the given namespaces.

        Returns:
            Dictionary containing FITS headers.
        """
        ...


__all__ = ["IFitsHeaderAfter"]



from abc import ABCMeta, abstractmethod
from typing import Tuple, Any

from .interface import Interface


class IOffsetsAltAz(Interface, metaclass=ABCMeta):
    """The module supports Alt/Az offsets, usually combined with :class:`~pyobs.interfaces.ITelescope` and
    :class:`~pyobs.interfaces.IPointingAltAz`."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def set_offsets_altaz(self, dalt: float, daz: float, **kwargs: Any) -> None:
        """Move an Alt/Az offset.

        Args:
            dalt: Altitude offset in degrees.
            daz: Azimuth offset in degrees.

        Raises:
            MoveError: If device could not be moved.
        """
        ...

    @abstractmethod
    async def get_offsets_altaz(self, **kwargs: Any) -> Tuple[float, float]:
        """Get Alt/Az offset.

        Returns:
            Tuple with alt and az offsets.
        """
        ...


__all__ = ["IOffsetsAltAz"]



import logging
from typing import Any

from pyobs.images.processor import ImageProcessor
from pyobs.images import Image
from pyobs.utils.fits import FilenameFormatter


log = logging.getLogger(__name__)


class CreateFilename(ImageProcessor):
    """Formats the filename for an image and stores it in FNAME."""

    __module__ = "pyobs.images.processors.misc"

    def __init__(self, pattern: str, **kwargs: Any):
        """Init an image processor that adds a filename to an image.

        Args:
            pattern: Filename pattern.
        """
        ImageProcessor.__init__(self, **kwargs)

        # default filename patterns
        if pattern is None:
            pattern = "{SITEID}{TELID}-{INSTRUME}-{DAY-OBS|date:}-{FRAMENUM|string:04d}-{IMAGETYP|type}01.fits"
        self._formatter = FilenameFormatter(pattern)

    async def __call__(self, image: Image) -> Image:
        """Add filename to image.

        Args:
            image: Image to add filename to.

        Returns:
            Image with filename in FNAME.
        """

        # copy image and set filename
        img = image.copy()
        img.format_filename(self._formatter)
        return img


__all__ = ["CreateFilename"]



import copy
import logging
from typing import List, Dict, Optional, Any
from astroplan import ObservingBlock
from astropy.time import TimeDelta
import astropy.units as u

from pyobs.robotic.task import Task
from pyobs.utils.time import Time
from .task import LcoTask
from .taskschedule import LcoTaskSchedule

log = logging.getLogger(__name__)


REQUEST = {
    "name": "Test",
    "observation_type": "NORMAL",
    "request": {
        "id": 1,
        "configurations": [
            {
                "constraints": {},
                "instrument_configs": [
                    {
                        "optical_elements": {},
                        "mode": None,
                        "exposure_time": 0.0,
                        "exposure_count": 1,
                        "bin_x": 3,
                        "bin_y": 3,
                    }
                ],
                "acquisition_config": {
                    "mode": "OFF",
                },
                "guiding_config": {
                    "mode": "OFF",
                },
                "target": {
                    "type": "ICRS",
                    "name": "bias",
                    "ra": 0.0,
                    "dec": 0.0,
                },
                "instrument_type": None,
                "type": "BIAS",
                "priority": 1,
                "configuration_status": {},
            }
        ],
        "windows": [],
        "duration": 10,
        "acceptability_threshold": 90.0,
    },
}


class LcoDummyTaskSchedule(LcoTaskSchedule):
    """Dummy scheduler for using the LCO portal"""

    def __init__(
        self,
        instrument_type: str,
        mode: str,
        binning: str,
        **kwargs: Any,
    ):
        """Creates a new LCO scheduler.

        Args:
            mode: Instrument mode
            instrument_type: Instrument type
        """
        LcoTaskSchedule.__init__(self, **kwargs)

        # set some stuff
        self._last_schedule_time = Time.now()

        # instruments
        self.instruments = {
            instrument_type.lower(): {
                "modes": {
                    "readout": {
                        "modes": [
                            {
                                "code": mode,
                                "name": mode,
                                "validation_schema": {"bin_x": {"default": binning}, "bin_y": {"default": binning}},
                            }
                        ]
                    }
                }
            }
        }

        # create task config
        cfg = copy.deepcopy(REQUEST)
        cfg["request"]["configurations"][0]["instrument_configs"][0]["mode"] = mode
        cfg["request"]["configurations"][0]["instrument_type"] = instrument_type
        cfg["start"] = Time.now()
        cfg["end"] = Time.now() + TimeDelta(5.0 * u.minute)

        # create task
        self._task: Optional[LcoTask] = self._create_task(LcoTask, config=cfg)

    async def _init_from_portal(self) -> None:
        pass

    async def last_scheduled(self) -> Optional[Time]:
        return self._last_schedule_time

    async def _update_schedule(self) -> None:
        pass

    async def update_now(self, force: bool = False) -> None:
        pass

    async def get_schedule(self, start_before: Time, end_after: Time, include_running: bool = True) -> Dict[str, Task]:
        return {} if self._task is None else {"task": self._task}

    def get_task(self, time: Time) -> Optional[LcoTask]:
        task = self._task
        self._task = None
        return task

    async def send_update(self, status_id: int, status: Dict[str, Any]) -> None:
        pass

    async def set_schedule(self, blocks: List[ObservingBlock], start_time: Time) -> None:
        pass


__all__ = ["LcoDummyTaskSchedule"]



from typing import Any
import pandas as pd
import logging

from .publisher import Publisher


log = logging.getLogger(__name__)


class CsvPublisher(Publisher):
    def __init__(self, filename: str, **kwargs: Any):
        """Initialize new CSV publisher.

        Args:
            filename: Name of file to log in.
        """
        Publisher.__init__(self, **kwargs)

        # store
        self._filename = filename

    async def __call__(self, **kwargs: Any) -> None:
        """Publish the given results.

        Args:
            **kwargs: Results to publish.
        """

        # load data
        csv = await self.data()

        # create new row from kwargs and append it
        row = pd.DataFrame(kwargs, index=[0])
        csv = pd.concat([csv, row], ignore_index=True)

        # write it
        await self.vfs.write_csv(self._filename, csv, index=False)

    async def data(self) -> pd.DataFrame:
        """Return data that has so far been published."""

        # load data
        try:
            # load it
            return await self.vfs.read_csv(self._filename, index_col=False)

        except FileNotFoundError:
            # file not found, so start new with row
            log.warning("No previous CSV file found, creating new one...")
            return pd.DataFrame()


__all__ = ["CsvPublisher"]



import logging
from typing import Tuple, List, Union, Dict, Any, Optional
import numpy as np
from numpy.typing import NDArray
from scipy import signal, optimize
from astropy.nddata import NDData
from astropy.table import Table, Column
import photutils

from pyobs.images import Image, ImageProcessor
from pyobs.mixins.pipeline import PipelineMixin
from .offsets import Offsets

log = logging.getLogger(__name__)


class CorrelationMaxCloseToBorderError(Exception):
    pass


class NStarOffsets(Offsets, PipelineMixin):
    """An offset-calculation method based on comparing 2D images of the surroundings of a variable number of stars."""

    def __init__(
        self,
        num_stars: int = 10,
        max_offset: float = 4.0,
        min_pixels: int = 3,
        min_sources: int = 1,
        pipeline: Optional[List[Union[Dict[str, Any], ImageProcessor]]] = None,
        **kwargs: Any,
    ):
        """Initializes a new auto guiding system.

        Requires pyobs.images.processors.detection.SepSourceDetection and
        pyobs.images.processors.photometry.SepPhotometry to be run on the image beforehand.

        Args:
            num_stars: maximum number of stars to use to calculate offset from boxes around them
            max_offset: the maximal expected offset in arc seconds. Determines the size of boxes
                around stars.
            min_pixels: minimum required number of pixels above threshold for source to be
                used for offset calculation.
            min_sources: Minimum required number of sources in image.
            pipeline: Pipeline to be used for first image in series.
        """
        Offsets.__init__(self, **kwargs)
        PipelineMixin.__init__(self, pipeline)

        # store
        self.num_stars = num_stars
        self.max_offset = max_offset
        self.min_pixels = min_pixels
        self.min_sources = min_sources
        self.ref_boxes: List[Any] = []

    def reset(self) -> None:
        """Resets guiding."""
        log.info("Reset auto-guiding.")
        self.ref_boxes = []

    def __call__(self, image: Image) -> Image:
        """Processes an image and sets x/y pixel offset to reference in offset attribute.

        Args:
            image: Image to process.

        Returns:
            Original image.

        Raises:
            ValueError: If offset could not be found.
        """

        # no reference image?
        if self.ref_boxes is None:
            log.info("Initialising nstar auto-guiding with new image...")
            star_box_size = max(5, self._get_box_size(self.max_offset, image.pixel_scale))
            log.info(f"Choosing box size of {star_box_size} pixels.")

            # initialize reference image information
            try:
                # get boxes
                self.ref_boxes = self._boxes_from_ref(image, star_box_size)

                # reset and finish
                image.meta["offsets"] = (0, 0)
                return image

            except ValueError as e:
                # didn't work
                log.warning(f"Could not initialize reference image info due to exception '{e}'. Resetting...")
                self.reset()
                if "offsets" in image.meta:
                    del image.meta["offsets"]
                self.offset = None, None
                return image

        # process it
        log.info("Perform auto-guiding on new image...")
        offsets = self._calculate_offsets(image)
        if offsets[0] is not None:
            image.meta["offsets"] = offsets
        return image

    @staticmethod
    def _get_box_size(max_expected_offset_in_arcsec, pixel_scale) -> int:
        # multiply by 4 to give enough space for fit of correlation around the peak on all sides
        return int(4 * max_expected_offset_in_arcsec / pixel_scale if pixel_scale else 20)

    async def _boxes_from_ref(self, image: Image, star_box_size: int) -> List:
        """Calculate the boxes around self.N_stars best sources in the image.

        Args:
             image: Image to process

        Returns:
            2-tuple with
                list of dimensions of boxes in "numpy" order: [0'th axis min, 0'th axis max, 1st axis min, 1st axis max]
                list of images of those boxes

        Raises:
            ValueError if not at least max(self.min_required_sources_in_image, self.N_stars) in filtered list of sources
        """

        # run pipeline on 1st image
        img = await self.run_pipeline(image)

        # check data and catalog
        if img.data is None:
            raise ValueError("No data found in image.")
        if img.catalog is None:
            raise ValueError("No catalog found in image.")

        # do photometry and get catalog
        sources = self._fits2numpy(img.catalog)

        # filter sources
        sources = self.remove_sources_close_to_border(sources, img.data.shape, star_box_size // 2 + 1)
        sources = self.remove_bad_sources(sources)
        self._check_sources_count(sources)
        selected_sources = self._select_brightest_sources(self.num_stars, sources)

        # extract boxes
        return photutils.psf.extract_stars(
            NDData(img.data.astype(float)), selected_sources, size=star_box_size
        ).all_stars

    @staticmethod
    def _fits2numpy(sources: Table) -> Table:
        """Convert from FITS to numpy conventions for pixel coordinates."""
        for k in ["x", "y", "xmin", "xmax", "ymin", "ymax", "xpeak", "ypeak"]:
            if k in sources:
                sources[k] -= 1
        return sources

    @staticmethod
    def remove_sources_close_to_border(sources: Table, image_shape: tuple, min_dist) -> Table:
        """Remove table rows from sources when source is closer than given distance from border of image.

        Args:
            sources: Input table.
            image_shape: Shape of image.
            min_dist: Minimum distance from border in pixels.

        Returns:
            Filtered table.
        ."""

        # get shape
        width, height = image_shape

        def min_distance_from_border(source) -> None:
            # calculate the minimum distance of source to any image border (across x and y)
            return min(width / 2 - np.abs(source["y"] - width / 2), height / 2 - np.abs(source["x"] - height / 2))

        sources.add_column(Column(name="min_dist", data=min_distance_from_border(sources)))
        sources.sort("min_dist")

        sources_result = sources[np.where(sources["min_dist"] > min_dist)]
        return sources_result

    def remove_bad_sources(
        self, sources: Table, max_ellipticity: float = 0.4, min_bkg_factor: float = 1.5, saturation: int = 50000
    ) -> Table:
        """Remove bad sources from table.

        Args:
            sources: Input table.
            max_ellipticity: Maximum ellipticity.
            min_bkg_factor: Minimum factor above local background.
            saturation: Saturation level.

        Returns:
            Filtered table.
        """

        # remove saturated sources
        sources = sources[sources["peak"] < saturation]

        # remove small sources
        sources = sources[np.where(sources["tnpix"] >= self.min_pixels)]

        # remove large sources
        tnpix_median = np.median(sources["tnpix"])
        tnpix_std = np.std(sources["tnpix"])
        sources = sources[np.where(sources["tnpix"] <= tnpix_median + 2 * tnpix_std)]

        # remove highly elliptic sources
        sources.sort("ellipticity")
        sources = sources[np.where(sources["ellipticity"] <= max_ellipticity)]

        # remove sources with background <= 0
        sources = sources[np.where(sources["background"] > 0)]

        # remove sources with low contrast to background
        sources = sources[np.where((sources["peak"] + sources["background"]) / sources["background"] > min_bkg_factor)]
        return sources

    @staticmethod
    def _select_brightest_sources(num_stars: int, sources: Table) -> Table:
        """Select N brightest sources from table.

        Args:
            num_stars: Maximum number of stars to select.
            sources: Source table.

        Returns:
            New table containing N brightest sources.
        """

        # sort by flux.
        sources.sort("flux")
        sources.reverse()

        # extract
        if 0 < num_stars < len(sources):
            sources = sources[:num_stars]
        return sources

    def _check_sources_count(self, sources: Table) -> None:
        """Check if enough sources in table.

        Args:
            sources: astropy table of sources to check.

        Returns:
            None

        Raises:
            ValueError if not at least self.min_sources in sources table

        """
        n_required_sources = self.min_sources
        if len(sources) < n_required_sources:
            raise ValueError(f"Only {len(sources)} source(s) in image, but at least {n_required_sources} required.")

    def _calculate_offsets(self, image: Image) -> Tuple[Optional[float], Optional[float]]:
        """Calculate offsets of given image to ref image for every star.

        Args:
            image: Image to calculate offset for.

        Returns:
            Offset tuple.
        """

        # data?
        if image.data is None:
            return None, None

        # calculate offset for each star
        offsets = []
        for box in self.ref_boxes:
            # get box dimensions
            box_ymin, box_ymax = box.origin[1], box.origin[1] + box.data.shape[0]
            box_xmin, box_xmax = box.origin[0], box.origin[0] + box.data.shape[1]

            # extract box image
            current_boxed_image = image.data[box_ymin:box_ymax, box_xmin:box_xmax].astype(float)

            # correlate
            corr = signal.correlate2d(current_boxed_image, box.data, mode="same", boundary="wrap")

            try:
                offset = self._offsets_from_corr(corr)
                offsets.append(offset)
            except Exception as e:
                log.info(f"Exception '{e}' caught. Ignoring this star.")
                pass

        if len(offsets) == 0:
            log.info(f"All {self.num_stars} fits on boxed star correlations failed.")
            return None, None
        offsets_np = np.array(offsets)
        return float(np.mean(offsets_np[:, 0])), float(np.mean(offsets_np[:, 1]))

    @staticmethod
    def _gauss2d(
        x: NDArray[float], a: float, b: float, x0: float, y0: float, sigma_x: float, sigma_y: float
    ) -> NDArray[float]:
        """2D Gaussian function."""
        return a + b * np.exp(-((x[0] - x0) ** 2) / (2 * sigma_x ** 2) - (x[1] - y0) ** 2 / (2 * sigma_y ** 2))

    def _offsets_from_corr(self, corr: NDArray[float]) -> Tuple[float, float]:
        """Fit 2d correlation data with a 2d gaussian + constant offset.
        raise CorrelationMaxCloseToBorderError if the correlation maximum is not well separated from border."""

        # check if maximum of correlation is too close to border
        self._check_corr_border(corr)

        # get x,y positions array corresponding to the independent variable values of the correlation
        x, y = NStarOffsets._corr_grid(corr)

        # shape data as needed by R^2 -> R scipy curve_fit
        xdata = np.vstack((x.ravel(), y.ravel()))
        ydata = corr.ravel()

        # estimate initial parameter values
        # constant offset of 2d gaussian
        a = np.min(corr)

        # height of 2d gaussian
        b = np.max(corr) - a

        # gaussian peak position (estimate from maximum pixel position in correlation)
        max_index = np.array(np.unravel_index(np.argmax(corr), corr.shape))
        x0, y0 = x[tuple(max_index)], y[tuple(max_index)]

        # estimate width of 2d gaussian as radius of area with values above half maximum
        half_max = np.max(corr - a) / 2 + a

        # sum over binary array
        greater_than_half_max_area = np.sum(corr >= half_max)
        sigma_x = np.sqrt(greater_than_half_max_area / np.pi)
        sigma_y = sigma_x

        # initial value list
        p0 = [a, b, x0, y0, sigma_x, sigma_y]
        bounds = (
            [-np.inf, -np.inf, x0 - sigma_x, y0 - sigma_y, 0, 0],
            [np.inf, np.inf, x0 + sigma_x, y0 + sigma_y, np.inf, np.inf],
        )

        # only use data points that clearly belong to peak to avoid border effects
        # mask_value_above_background = ydata > -1e5  # a + .1*b
        mask_circle_around_peak = (x.ravel() - x0) ** 2 + (y.ravel() - y0) ** 2 < 4 * (sigma_x ** 2 + sigma_y ** 2) / 2
        mask = mask_circle_around_peak
        ydata_restricted = ydata[mask]
        xdata_restricted = xdata[:, mask]

        # do fit
        try:
            popt, pcov = optimize.curve_fit(self._gauss2d, xdata_restricted, ydata_restricted, p0, bounds=bounds)
        except Exception as e:
            # if fit fails return max pixel
            log.info("Returning pixel position with maximal value in correlation.")
            idx = np.unravel_index(np.argmax(corr), corr.shape)
            return float(idx[0]), float(idx[1])

        # check quality of fit
        median_squared_relative_residue_threshold = 1e-1
        fit_ydata_restricted = self._gauss2d(xdata_restricted, *popt)
        square_rel_res = np.square((fit_ydata_restricted - ydata_restricted) / fit_ydata_restricted)
        median_squared_rel_res = np.median(np.square(square_rel_res))

        if median_squared_rel_res > median_squared_relative_residue_threshold:
            raise Exception(
                f"Bad fit with median squared relative residue = {median_squared_rel_res}"
                f" vs allowed value of {median_squared_relative_residue_threshold}"
            )

        return popt[2], popt[3]

    @staticmethod
    def _corr_grid(corr: NDArray[float]) -> NDArray[float]:
        """Create x/y grid for given 2D correlation."""
        xs = np.arange(-corr.shape[0] / 2, corr.shape[0] / 2) + 0.5
        ys = np.arange(-corr.shape[1] / 2, corr.shape[1] / 2) + 0.5
        return np.meshgrid(xs, ys)

    @staticmethod
    def _check_corr_border(corr: NDArray[float]) -> None:
        """Check whether maximum of correlation is too close to border."""

        corr_size = corr.shape[0]
        x, y = NStarOffsets._corr_grid(corr)

        max_index = np.array(np.unravel_index(np.argmax(corr), corr.shape))
        x0, y0 = x[tuple(max_index)], y[tuple(max_index)]

        if x0 < -corr_size / 4 or x0 > corr_size / 4 or y0 < -corr_size / 4 or y0 > corr_size / 4:
            raise CorrelationMaxCloseToBorderError(
                "Maximum of correlation is outside center half of axes. "
                "This means that either the given image data is bad, or the offset is larger than expected."
            )


__all__ = ["NStarOffsets"]
